/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Label.h"
#include "LabelSettings.h"
#include "xFont.h"
#include "Texture.h"
#include "SystemInfo.h"
#include "GhostLabel.h"

#include "shlobj.h"

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

//MouseWheel
//-----------------------------------------------------
#ifndef WM_MOUSEWHEEL
	#define WM_MOUSEWHEEL 0x020A
#endif
//-----------------------------------------------------

//Ballon Tooltip
//-----------------------------------------------------
#ifndef TTS_BALLOON
	#define TTS_BALLOON 0x40
#endif
//-----------------------------------------------------

//Text Alignment
//-----------------------------------------------------
#define ALIGN_LEFT 0
#define ALIGN_TOP 0
#define ALIGN_CENTER 1
#define ALIGN_RIGHT 2
#define ALIGN_BOTTOM 2
//-----------------------------------------------------

//Timers
//-----------------------------------------------------
#define TIMER_MOUSETRACK 1
#define TIMER_UPDATE 2
#define TIMER_SCROLL 3
#define TIMER_AUTOHIDE 4
#define TIMER_ALPHAFADEIN 5
#define TIMER_ALPHAFADEOUT 6
#define TIMER_FRAMEANIMATION 7
#define TIMER_SCROLLPAUSE 8
//-----------------------------------------------------

//SendMessage
//-----------------------------------------------------
#define LM_SETLABELTEXT 9600
#define LM_SETCURSOR    9610
#define LM_EXTERNALDRAW 9630
#define LM_EXTERNALDRAWNOTIFY 9640
#define LM_TEXTEDITBOX 9650
//-----------------------------------------------------

//Alpha Transparency
//-----------------------------------------------------
#ifndef WS_EX_LAYERED
	#define WS_EX_LAYERED           0x00080000L
	#define LWA_COLORKEY            0x00000001L
	#define ULW_COLORKEY            0x00000001L
	#define LWA_ALPHA               0x00000002L
	#define ULW_ALPHA               0x00000002L
	#define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

#ifndef AC_SRC_ALPHA
	#define AC_SRC_ALPHA            0x00000001L
#endif

typedef BOOL (WINAPI *PULW)(HWND hwnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
typedef DWORD (WINAPI *PSLWA)(HWND, DWORD, BYTE, DWORD);

static PULW pUpdateLayeredWindow = NULL;
static PSLWA pSetLayeredWindowAttributes = NULL;

static BOOL alphaInitialized = FALSE;
//-----------------------------------------------------

//Text EditBox
//-----------------------------------------------------
static WNDPROC wpDefault;

BOOL CALLBACK EditProc(HWND hText, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char buf[MAX_PATH] = {0};

	switch(msg)
	{
		case WM_CHAR:
		{
			if( wParam == VK_RETURN )
			{
				GetWindowText(hText, buf, sizeof(buf));

				SendMessage(GetParent(hText), LM_TEXTEDITBOX, 0, (long) buf);
			}
			else if ( wParam == VK_ESCAPE )
			{
				SetWindowText(hText,"");
				ShowWindow(hText, SW_HIDE);
			}
		}

		break;

		case WM_KILLFOCUS:

			SendMessage(GetParent(hText), LM_TEXTEDITBOX, 0, (long) "");

		break;

    }

	return CallWindowProc(wpDefault, hText, msg, wParam, lParam);
}
//-----------------------------------------------------

extern int screenX;
extern int screenY;

extern int bitsperpixel;

extern LabelList labelList;

Label::Label(const string &name):actualheight(0),actualwidth(0),actualx(0),actualy(0)
{
	this->name = name;

	hWnd = 0;
	hInstance = 0;
	box = 0;

	hWndTextBox = 0;

	sender = 0;

	TooltipHints = 0;
	cursor = 0;
	handcursor = 0;

	labelInitialized = false;

	background = 0;
	font = 0;

	borderMethod = 0;
	hoverborderMethod = 0;
	pressedborderMethod = 0;

	hasHookedModules = false;
	parentIsAlphaMapped = false;

	backgroundDC = 0;
	backgroundBitmap = 0;
	backupDC = 0;
	backupBitmap = 0;

	hoverActive = false;
	pressedActive = false;

	useHover = false;
	usePressed = false;

	mousePressed = false;
	mouseInside = false;

	mouseInsideHREF = false;

	movemodifierkeyPressed = false;
	moveButtonPressed = 0;

	hasMoved = false;

	dockedtoSide = -1;

	visible = false;

	bUsingDefSkin = false;

	dynamicText = false;
	dynamicTooltip = false;

	text = "";
	originalText.clear();
	originalTextBackup = "";
	hrefconfig.clear();
	current = 0;
	currentBackup = 0;

	scroll = 0;
	scrollBackup = 1;
	scrollLimit = 0;
	scrollPosition = 0;
	frozen = false;

	scrollPause = 0;
	scrollSteps = 0;
	scrollActualStep = 1;
	scrollStepsInitial = 0;
	
	//Animations
	frameCounter = 0;
	loopCounter = -1;
	normalAnim = false;
	hoverAnim = false;

	layeredFlag = false;
	UpdateAlpha = 0;

	alphaFadeStep = 2;
	alphaFadeDelay = 10;

	useresizeBorder = false;
	resizeBorderLeft = -1;
	resizeBorderRight = -1;
	resizeBorderTop = -1;
	resizeBorderBottom = -1;

	useTextFade = false;
	textFadeLeft = -1;
	textFadeRight = -1;
	textFadeTop = -1;
	textFadeBottom = -1;

	//TextEdit Box
	textboxXOrg = "";
	textboxYOrg = "";
	textboxWidthOrg = "";
	textboxStartHidden = false;
	textboxAutoHide = false;
	textboxAlign = 0;
	textboxVisible = false;

	labelLeftClickRegions.clear();
	labelRightClickRegions.clear();
	labelMiddleClickRegions.clear();

	labelHREFClickRegions.clear();

	snapToList.clear();

	dockedToList.clear();
	hasDockedLabelList.clear();

	ghostlabelList.clear();
}

Label::~Label()
{
	if (cursor && !cursorPath.empty())
		DestroyCursor(cursor);
	if (handcursor)
		DestroyCursor(handcursor);

	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
		TooltipHints = 0;
	}

	if (hWnd && hWndTextBox)
	{
		DestroyWindow(hWndTextBox);
		hWndTextBox = 0;
	}

	for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
		delete *glit;

	ghostlabelList.clear();

	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_SCROLL);
		KillTimer(hWnd, TIMER_AUTOHIDE);
		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);
		KillTimer(hWnd, TIMER_FRAMEANIMATION);
		KillTimer(hWnd, TIMER_SCROLLPAUSE);

		DestroyWindow(hWnd);
	}

	if(backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
	}
	if (backgroundDC)
		DeleteDC(backgroundDC);

	if(backupBitmap)
	{
		backupBitmap = (HBITMAP) SelectObject(backupDC, backupBitmap);
		DeleteObject(backupBitmap);
	}
	if (backupDC)
		DeleteDC(backupDC);

	if( background && !bUsingDefSkin  )
		delete background;

	if( font && font != defaultSettings->font )
		delete font;
}

void Label::load(HINSTANCE hInstance, HWND box, bool dropShadow)
{
	this->hInstance = hInstance;
	
	if (IsWindow(box))
	{
		this->box = box;
	}
	else
	{
		this->box = NULL;
	}
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	if(!IsOS(OS_XP))
		dropShadow = false;

	hWnd = CreateWindowEx( box ? 0 : WS_EX_TOOLWINDOW,
			(dropShadow ? "xLabelXP" : "xLabel"),
			name.c_str(),
			(box ? WS_CHILD : WS_POPUP) | WS_CLIPCHILDREN,
			actualx, actualy,		
			actualwidth, actualheight,
			box ? box : GetLitestepWnd(), 
			0,
			hInstance,
			this);
		
	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, magicDWord);

		reconfigure();

		if ( parentIsAlphaMapped )
			SendMessage(hWnd, WM_PAINT, 0, 0);
	}
	else
		delete this;
}

void Label::reconfigure(bool refresh)
{
	labelInitialized = true;

	//Initialize Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";

	LabelSettings settings(name.c_str(), box);
	
	//Extra Settings
	labelGroup = settings.labelGroup;
	bUseFahrenheit = settings.bUseFahrenheit;

	//Cursor
	cursorPath = settings.cursorPath;

	//Module Hook
	if (!refresh)
		moduleHook = GetxRCLineList( ("*" + name + "ModuleHook") );

	//Label Docking and ExtraImage
	if (refresh)
	{
		dockedToList.clear();
		hasDockedLabelList.clear();
		
		for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			delete *glit;

		ghostlabelList.clear();
	}
	dockedToList = GetxRCLineList( ("*" + name + "DockedTo") );

	dockOnSnapTo = settings.dockOnSnapTo;

	numTokens = SplitString(settings.behindWindow, one, two, three, four, five);

	if ( numTokens >= 1 )
	{
		afterClass = one;
		if ( numTokens == 2 )
			afterTitle = two;
	}

	//TextEdit Box
	numTokens = SplitString(settings.texteditBox, one, two, three, four, five);

	if ( numTokens >= 3 )
	{
		textboxXOrg = one;
		textboxYOrg = two;
		textboxWidthOrg = three;

		textboxStartHidden = strchr(four.c_str(), 'h') != 0;
		textboxAutoHide = strchr(four.c_str(), 'a') != 0;
		if ( strchr(four.c_str(), 'r') != 0 )
			textboxAlign = 2;
		else if ( strchr(four.c_str(), 'c') != 0 )
			textboxAlign = 1;
		else
			textboxAlign = 0;
	}

	enterAction = settings.enterAction;

	//Painting Background
	//Use Split Helper Strings
	numTokens = SplitString(settings.borderMethod, one, two, three, four, five);

	if(numTokens >= 1)
	{	
		if (stricmp(one.c_str(), "raised") == NULL)
			borderMethod = 1;
		else if (stricmp(one.c_str(), "sunken") == NULL)
			borderMethod = 2;
		else if (stricmp(one.c_str(), "etched") == NULL)
			borderMethod = 3;
		else if (stricmp(one.c_str(), "bump") == NULL)
			borderMethod = 4;
		else 
			borderMethod = 0;
	}
	else
		borderMethod = 0;

	hoverborderMethod = borderMethod;
	pressedborderMethod = borderMethod;

	if (numTokens >= 2)
	{
		if (stricmp(two.c_str(), "raised") == NULL)
			hoverborderMethod = 1;
		else if (stricmp(two.c_str(), "sunken") == NULL)
			hoverborderMethod = 2;
		else if (stricmp(two.c_str(), "etched") == NULL)
			hoverborderMethod = 3;
		else if (stricmp(two.c_str(), "bump") == NULL)
			hoverborderMethod = 4;
		else if (stricmp(two.c_str(), ".none") == NULL)
			hoverborderMethod = 0;
		else 
			hoverborderMethod = borderMethod;
	}
	if (numTokens == 3)
	{
		if (stricmp(three.c_str(), "raised") == NULL)
			pressedborderMethod = 1;
		else if (stricmp(three.c_str(), "sunken") == NULL)
			pressedborderMethod = 2;
		else if (stricmp(three.c_str(), "etched") == NULL)
			pressedborderMethod = 3;
		else if (stricmp(three.c_str(), "bump") == NULL)
			pressedborderMethod = 4;
		else if (stricmp(three.c_str(), ".none") == NULL)
			pressedborderMethod = 0;
		else 
			pressedborderMethod = hoverborderMethod;
	}

	setBackground(settings.skin);	
	transparencyNeeded = background->isTransparencyNeeded();

	//Transparency Settings
	if ( box && transparencyNeeded )
		transparencyMode = 1;
	else if ( transparencyNeeded )
		transparencyMode = settings.transparencyMode;
	else 
		transparencyMode = 2;

	alphaTransparency = settings.alphaTransparency;
	alphaMap = settings.alphaMap;
	alphaFade = settings.alphaFade;

	//Use Split Helper Strings
	numTokens = SplitString(settings.customalphaFade, one, two, three, four, five);
	if (numTokens >= 1)
	{
		alphaFadeStep = atoi( one.c_str() );
		if (alphaFadeStep < 1 || alphaFadeStep > 255)
			alphaFadeStep = 2;
	}
	if  (numTokens == 2)
	{
		alphaFadeDelay = atoi( two.c_str() );
		if (alphaFadeDelay < 1 || alphaFadeDelay > 500)
			alphaFadeDelay = 10;
	}

	if ( background->isIconUsed() && IsOS( OS_XP ) )
		alphaMap = true;

	if ( (alphaTransparency != 255 || alphaMap || alphaFade) && IsOS(OS_2KXP) )
	{
		if ( (GetxRCString(name.c_str(), "Image", "")).empty() && (GetxRCString(labelGroup.c_str(), "Image", "")).empty() && (GetxRCString("AllLabels", "Image", "")).empty())
			alphaMap = false;
	}
	else
	{
		alphaTransparency = 255;
		alphaMap = false;
		alphaFade = false;
	}

	//AutoSize
	autoWidthMode = settings.autoWidthMode;
	autoHeightMode = settings.autoHeightMode;
	autoMinWidth = settings.autoMinWidth;
	autoMinHeight = settings.autoMinHeight;
	autoMaxWidth = settings.autoMaxWidth;
	autoMaxHeight = settings.autoMaxHeight;
	if (!refresh)
	{
		startCenterX = settings.x+(settings.width/2);
		startCenterY = settings.y+(settings.height/2);
	}

	//Hover/Pressed Images
	if ( !(GetxRCString(name.c_str(), "HoverImage", "")).empty() || !(GetxRCString(labelGroup.c_str(), "HoverImage", "")).empty() || !(GetxRCString("AllLabels", "HoverImage", "")).empty() ||
			!(GetxRCString(name.c_str(), "HoverFrames", "")).empty() || !(GetxRCString(labelGroup.c_str(), "HoverFrames", "")).empty() || !(GetxRCString("AllLabels", "HoverFrames", "")).empty() ||
			(hoverborderMethod != 0 && hoverborderMethod != borderMethod) || 
			!(GetxRCString(name.c_str(), "HoverSolidColors", "")).empty() || !(GetxRCString(labelGroup.c_str(), "HoverSolidColors", "")).empty() || !(GetxRCString("AllLabels", "HoverSolidColors", "")).empty() )
		useHover = true;
	
	if ( !(GetxRCString(name.c_str(), "PressedImage", "")).empty() || !(GetxRCString(labelGroup.c_str(), "PressedImage", "")).empty() || !(GetxRCString("AllLabels", "PressedImage", "")).empty() ||
			(pressedborderMethod != 0 && pressedborderMethod != hoverborderMethod && pressedborderMethod != borderMethod) || 
			!(GetxRCString(name.c_str(), "PressedSolidColors", "")).empty() || !(GetxRCString(labelGroup.c_str(), "PressedSolidColors", "")).empty() || !(GetxRCString("AllLabels", "PressedSolidColors", "")).empty())
		usePressed = true;

	//Use Split Helper Strings
	numTokens = SplitString(settings.resizeBorder, one, two, three, four, five);
	if (numTokens == 1)
	{
		resizeBorderLeft = resizeBorderRight = resizeBorderTop = resizeBorderBottom = atoi(one.c_str());
	}
	else if (numTokens == 2)
	{
		resizeBorderLeft = resizeBorderRight = atoi(one.c_str());
		resizeBorderTop = resizeBorderBottom = atoi(two.c_str());
	}
	else if (numTokens == 4)
	{
		resizeBorderLeft = atoi(one.c_str());
		resizeBorderRight = atoi(two.c_str());
		resizeBorderTop = atoi(three.c_str());
		resizeBorderBottom = atoi(four.c_str());	
	}

	if ( resizeBorderLeft <= 0 )
		resizeBorderLeft = -1;
	if ( resizeBorderRight <= 0 )
		resizeBorderRight = -1;
	if ( resizeBorderTop <= 0 )
		resizeBorderTop = -1;
	if ( resizeBorderBottom <= 0 )
		resizeBorderBottom = -1;

	if ( resizeBorderLeft > -1 || resizeBorderRight > -1 || resizeBorderTop > -1 || resizeBorderBottom > -1 )
		useresizeBorder = true;

	//Normal Animation
	imageLoop = GetxRCInt(name, "FrameLoops", -1);
	imageFrameDelay = GetxRCInt(name, "FrameDelay", 100, 1);
	imageFrameCount = GetxRCInt(name, "FrameCount", 0, 0);
	if (imageFrameCount != 0)
		normalAnim = true;
	
	//Hover Animation
	hoverimageLoop = GetxRCInt(name, "HoverFrameLoops", -1);
	hoverimageFrameDelay = GetxRCInt(name, "HoverFrameDelay", 100, 1);
	hoverimageFrameCount = GetxRCInt(name, "HoverFrameCount", 0, 0);
	if (hoverimageFrameCount != 0)
		hoverAnim = true;

	//Painting Text
	setFont(settings.font);	
	align = settings.align;
	vertAlign = settings.vertAlign;
	leftBorder = settings.leftBorder;
	topBorder = settings.topBorder;
	rightBorder = settings.rightBorder;
	bottomBorder = settings.bottomBorder;
	noEscape = settings.noEscape;
	//Must set BEFORE setText to work
	noTextSplit = settings.noTextSplit;
	if (!refresh)
		setText(settings.text, true);

	hrefconfig = split(settings.hrefconfig, ";");

	//Use Split Helper Strings
	if ( autoWidthMode == 0 && autoHeightMode == 0 )
	{
		numTokens = SplitString(settings.textFade, one, two, three, four, five);
		if (numTokens == 1)
		{
			textFadeLeft = textFadeRight = textFadeTop = textFadeBottom = atoi(one.c_str());
		}
		else if (numTokens == 2)
		{
			textFadeLeft = textFadeRight = atoi(one.c_str());
			textFadeTop = textFadeBottom = atoi(two.c_str());
		}
		else if (numTokens == 4)
		{
			textFadeLeft = atoi(one.c_str());
			textFadeRight = atoi(two.c_str());
			textFadeTop = atoi(three.c_str());
			textFadeBottom = atoi(four.c_str());
		}

		if ( textFadeLeft <= 0 )
			textFadeLeft = 0;

		if ( textFadeRight <= 0 )
			textFadeRight = 0;

		if ( textFadeTop <= 0 )
			textFadeTop = 0;

		if ( textFadeBottom <= 0 )
			textFadeBottom = 0;

		if ( textFadeLeft > 0 || textFadeRight > 0 || textFadeTop > 0 || textFadeBottom > 0 )
			useTextFade = true;
	}
	
	//Scrolling
	scrollPadLength = settings.scrollPadLength;
	scrollInterval = settings.scrollInterval;
	scrollSpeed = settings.scrollSpeed;
	if (!refresh)
		setScrolling(settings.scroll);

	if ( scroll == 2 || scroll == 4 )
	{
		//Use Split Helper Strings
		numTokens = SplitString(settings.scrollPause, one, two, three, four, five);
		if (numTokens == 3)
		{
			scrollStepsInitial = atoi(one.c_str());
			scrollSteps = atoi(two.c_str());
			scrollPause = atoi(three.c_str());
		}
	}

	//Updates
	alwaysUpdateContent = settings.alwaysUpdateContent;
	updateInterval = settings.updateInterval;

	//Events/Actions
	leftClickCommand = settings.leftClickCommand;
	leftClickDownCommand = settings.leftClickDownCommand;
	leftClickUpCommand = settings.leftClickUpCommand;
	leftDoubleClickCommand = settings.leftDoubleClickCommand;
	middleClickCommand = settings.middleClickCommand;
	middleClickDownCommand = settings.middleClickDownCommand;
	middleClickUpCommand = settings.middleClickUpCommand;
	middleDoubleClickCommand = settings.middleDoubleClickCommand;
	rightClickCommand = settings.rightClickCommand;
	rightClickDownCommand = settings.rightClickDownCommand;
	rightClickUpCommand = settings.rightClickUpCommand;
	rightDoubleClickCommand = settings.rightDoubleClickCommand;
	wheelDownCommand = settings.wheelDownCommand;
	wheelUpCommand = settings.wheelUpCommand;
	enterCommand = settings.enterCommand;
	leaveCommand = settings.leaveCommand;
	dropCommand = settings.dropCommand;
	textChangeCommand = settings.textChangeCommand;
	resizeCommand = settings.resizeCommand;
	moveCommand = settings.moveCommand;
	hideCommand = settings.hideCommand;
	showCommand = settings.showCommand;

	fireOnMoveDuringMove = settings.fireOnMoveDuringMove;

	//Event Regions
	labelLeftClickRegions = GetxRCLineList( ("*" + name + "OnLeftClickRegion").c_str() );
	labelRightClickRegions = GetxRCLineList( ("*" + name + "OnRightClickRegion").c_str() );
	labelMiddleClickRegions = GetxRCLineList( ("*" + name + "OnMiddleClickRegion").c_str() );

	//Snap To Settings
	//Use Split Helper Strings
	numTokens = SplitString(settings.snapTo, one, two, three, four, five);
	snapDistance = atoi( one.c_str() );
	if (numTokens == 1)
		snapTo = "desktop";
	else
		snapTo = two;

	if (settings.focusOnEnter)
		focusOnEnter = 1;
	else
		focusOnEnter = 0;

	//Moving
	moveable = settings.moveable;
	moveKey = settings.moveKey;
	moveButton = settings.moveButton;

	if (cursor)
	{
		DestroyCursor(cursor);
		cursor = 0;
	}
		
	SendMessage(hWnd, LM_SETCURSOR, 0, 0);

	if ( IsOS(OS_2KXP) )
		ghosted = settings.ghosted;
	else
		ghosted = false;

	if( alphaTransparency != 255 || alphaMap || alphaFade || ghosted )
	{
		if (!refresh)
		{
			// MUST BE SET BEFORE setAlwaysOnTop()
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);
				
			setGhosted(ghosted);

			layeredFlag = true;

			if (alphaFade)
				UpdateLabelAlpha(0);
			else
				UpdateLabelAlpha(alphaTransparency);
		}
		else if (refresh)
			setGhosted(ghosted);
	}

	//File Actions
	if (dropCommand.empty())
		DragAcceptFiles(hWnd, false);
	else
		DragAcceptFiles(hWnd, true);

	//AlwaysOnTop
	if (!refresh)
		setAlwaysOnTop(settings.alwaysOnTop);

	startHidden = settings.startHidden;

	//Initial Repositioning
	if (!refresh)
		reposition(3, settings.x, settings.y, settings.width, settings.height);

	if ( refresh )
	{
		for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			delete *glit;

		ghostlabelList.clear();
	}

	StringList extraPainting = GetxRCLineList( ("*" + name + "GhostImage").c_str() );
	//Extra Painting Background
	
	for(StringListIterator sit = extraPainting.begin(); sit != extraPainting.end(); sit++)
	{
		//Use Split Helper Strings
		numTokens = SplitString((*sit), one, two, three, four, five);
		
		if ( numTokens >= 3 )
		{
			GhostLabel* newimage = new GhostLabel();
		
			newimage->orgx = one;
			newimage->orgy = two;
			
			newimage->x = xParseCoord(one, 0, actualwidth);
			newimage->y = xParseCoord(two, 0, actualheight);
			
			HDC tmpDC = CreateCompatibleDC(NULL);
			HBITMAP tmpBitmap = LoadLSImage(three.c_str(), 0);
			SelectObject(tmpDC, tmpBitmap);

			BITMAP bm;
			GetObject(tmpBitmap, sizeof(BITMAP), &bm);

			GetLSBitmapSize(tmpBitmap, &newimage->width, &newimage->height);

			newimage->memoryDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=newimage->width;      // Set size you need
			alpha.bmiHeader.biHeight=newimage->height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			newimage->memoryBitmap = CreateDIBSection( newimage->memoryDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

			SelectObject(newimage->memoryDC, newimage->memoryBitmap);

			BitBlt(newimage->memoryDC, 0, 0, newimage->width, newimage->height, tmpDC, 0, 0, SRCCOPY);

			if(tmpBitmap)
			{
				tmpBitmap = (HBITMAP) SelectObject(tmpDC, tmpBitmap);
				DeleteObject(tmpBitmap);
			}
			if (tmpDC)
				DeleteDC(tmpDC);

			if ( numTokens >= 4 && four[0] == '#' )
			{
				newimage->alphamap = strchr(four.c_str(), 'a') != 0;
				newimage->normal = strchr(four.c_str(), 'n') != 0;
				newimage->hover = strchr(four.c_str(), 'h') != 0;
				newimage->pressed = strchr(four.c_str(), 'p') != 0;
				newimage->fullstretch = strchr(four.c_str(), 'f') != 0;
			}
			if ( numTokens == 5 )
			{
				int tmpalpha = atoi( five.c_str() );
				if (tmpalpha < 0 || tmpalpha > 255)
					newimage->alpha = 255;
				else
					newimage->alpha = tmpalpha;
			}

			//Secure Functions
			//------------------------------------------------------------------------
			if ( !newimage->normal && !newimage->hover && !newimage->pressed )
			{
				newimage->normal = true;
				newimage->hover = true;
				newimage->pressed = true;
			}
			if ( newimage->fullstretch && (newimage->x != 0 || newimage->y != 0) )
				newimage->fullstretch = false;
			//------------------------------------------------------------------------

			if ( newimage->alphamap || newimage->alpha != 255 )
			{
				GetObject(newimage->memoryBitmap, sizeof(BITMAP), &bm);

				if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
				{
					BYTE *pPixel = (LPBYTE)bm.bmBits;

					for (int i = 0; i < bm.bmHeight * bm.bmWidth; ++i)
					{	
						if ( !newimage->alphamap && newimage->alpha != 255 )
						{
							//Prepare
							if (pPixel[3] == 0)
								pPixel[3] = 255;
						}

						//Magic Pink
						if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
							pPixel[3] = 0;

						//PreMultiply
						pPixel[0] = pPixel[0] * pPixel[3] / 255;
						pPixel[1] = pPixel[1] * pPixel[3] / 255;
						pPixel[2] = pPixel[2] * pPixel[3] / 255;
						
						pPixel+= 4;
					}
				}
			}

			if ( newimage->memoryBitmap != 0 )
				ghostlabelList.insert(ghostlabelList.end(), newimage);
			else
				delete newimage;
		}
	}

	//ToolTips
	balloontooltip = settings.balloontooltip;
	originalTooltip = settings.tooltip;

	tooltipmaxwidth = settings.tooltipmaxwidth;
	if (!originalTooltip.empty())
	{
		while ( strstr(originalTooltip.c_str(), "\\n") != NULL )
		{
			originalTooltip.replace(originalTooltip.find("\\n"), 2, "\n", 1);
		}
		while ( strstr(originalTooltip.c_str(), "<br>") != NULL )
		{
			originalTooltip.replace(originalTooltip.find("<br>"), 4, "\n", 1);
		}

		if (TooltipHints == 0)
		{
			TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP | (balloontooltip ? TTS_BALLOON : 0),
										  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
										  NULL, NULL, hInstance, NULL);
			if (TooltipHints != 0)
			{
				SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
							 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

				SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, tooltipmaxwidth);

				strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());
				addHint(hWnd, tooltip);
			}
		}
		else
		{
			strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());
			updateHint(hWnd, tooltip);
		}
	}
	else
	{
		if (hWnd && TooltipHints != 0)
		{
			removeHint(hWnd);
			DestroyWindow(TooltipHints);
			TooltipHints = 0;
		}
	}
	
	if ( !hWndTextBox && ( !textboxXOrg.empty() || !textboxYOrg.empty() || !textboxWidthOrg.empty()) && !alphaMap )
	{
		hWndTextBox = CreateWindowEx(WS_EX_CLIENTEDGE,
							"EDIT",
							"",
							WS_CHILD | ES_AUTOHSCROLL | (textboxAlign == 1 ? ES_CENTER : 0) | (textboxAlign == 2 ? ES_RIGHT : 0),
							0, 0,
							0, 0,
							hWnd,
							0,
							hInstance,
							0);

	}
	if ( hWndTextBox )
	{
		SendMessage(hWndTextBox, WM_SETFONT, (WPARAM) font->gethFont(), FALSE);

		RECT r;

		r.left = 0;
		r.top = 0;
		r.right = 0;
		r.bottom = 0;

		HFONT hFont = (HFONT) SelectObject(backgroundDC, font->gethFont());
				
		string testtext = "Test Text Calculation!";
		DrawTextEx(backgroundDC, (char *) testtext.c_str(), testtext.length(), &r, DT_SINGLELINE | DT_CALCRECT, 0);
		textboxNeededHeight = (r.bottom-r.top)+6;

		SelectObject(backgroundDC, hFont);

		SetWindowPos(hWndTextBox, NULL, 
						xParseCoord(textboxXOrg, 0, actualwidth),
						xParseCoord(textboxYOrg, 0, actualheight),
						xParseDimen(textboxWidthOrg, actualwidth, actualwidth), 
						textboxNeededHeight,
						SWP_NOACTIVATE | SWP_NOZORDER);
			

		if ( !refresh )
		{
			wpDefault = (WNDPROC) SetWindowLong(hWndTextBox, GWL_WNDPROC, (long) EditProc);
				
			if ( !textboxStartHidden )
			{
				ShowWindow(hWndTextBox, SW_SHOW);
				textboxVisible = true;
			}
		}
	}
	else
	{
		textboxXOrg = textboxYOrg = textboxWidthOrg = "";
		hWndTextBox = 0;
	}

	update();

	if (normalAnim)
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
}

//------------------------------------------------------------
//Set Background
//------------------------------------------------------------

void Label::setBackground(Texture *newbackground)
{
	if(!bUsingDefSkin)
		delete background;

	background = newbackground;

	bUsingDefSkin = (background == defaultSettings->skin);	
}

//------------------------------------------------------------
//Set Parent Box (nearly never needed/called!)
//------------------------------------------------------------

void Label::setBox(HWND newparent)
{
	if (box == newparent)
		return;

	ModifyStyleEx(hWnd, WS_EX_LAYERED, 0);

	if (alwaysOnTop)
		setAlwaysOnTop(FALSE);
	
	box = newparent;
	
	ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
	SetParent(hWnd, newparent);

	transparencyMode = 1;
}

//------------------------------------------------------------
//"Modify Styles"
//------------------------------------------------------------

void Label::setAlwaysOnTop(bool alwaysOnTop)
{
	if (box)
		return;
	
	this->alwaysOnTop = alwaysOnTop;
	
	ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
	SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
	ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
		
	SetWindowPos(hWnd, alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
			0, 0, 0, 0,
			SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
}

void Label::setAlpha(int newAlphaTransparency, bool updateValue)
{
	if (box)
		return;

	if (newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
	{
		if (!layeredFlag)
		{
			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, 0);
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			layeredFlag = true;
		}
		
		if (updateValue)
		{
			alphaTransparency = newAlphaTransparency;
		}
		
		UpdateLabelAlpha(newAlphaTransparency);
	}
}

void Label::setGhosted(bool ghosted)
{	
	if (box)
		return;

	this->ghosted = ghosted;
	
	if (ghosted)
		ModifyStyleEx(hWnd, 0, WS_EX_TRANSPARENT);
	else
		ModifyStyleEx(hWnd, WS_EX_TRANSPARENT, 0);
}

//------------------------------------------------------------
//Set Font
//------------------------------------------------------------

void Label::setFont(xFont *newfont)
{
	if(font != defaultSettings->font)
		delete font;

	font = newfont;
}

//------------------------------------------------------------
//Set Text Settings
//------------------------------------------------------------

void Label::setText(const string &text, bool updateBackup, bool textEditBox)
{
	if ( textEditBox )
	{
		if ( hWndTextBox )
		{
			SetWindowText( hWndTextBox, (systemInfo->processLabelText(text, this, NULL)).c_str() );
			if ( !textboxVisible && visible )
				ShowWindow(hWndTextBox, SW_SHOW);
		}
	}
	else
	{
		//Backup User Changed TEXT <-> Internal Change
		if ( updateBackup )
		{
			originalTextBackup = text;
			labelHREFClickRegions.clear();
		}
		
		string process = text;
		if (!process.empty())
		{
			while ( strstr(process.c_str(), "\\n") != NULL )
				process.replace(process.find("\\n"), 2, "<br>", 4);
		}

		if (noTextSplit)
			originalText = split(process, "%&*&%");
		else
			originalText = split(process, ";");

		if ( updateBackup )
		{
			current = 0;
			currentBackup = 0;
		}
		else
		{
			current = currentBackup;
		}

		update();
	}
}

void Label::previous()
{
	current--;

	if(current < 0)
		current = originalText.size() - 1;

	currentBackup = current;

	update();
}

void Label::next()
{
	current++;

	if(current >= originalText.size())
		current = 0;

	currentBackup = current;

	update();
}

void Label::setScrolling(int scrolling)
{
	if (scrolling != 0 && scroll == 0)
	{
		scrollPosition = 0;
		SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
	}
	else if (scrolling == 0 && scroll != 0)
	{
		KillTimer(hWnd, TIMER_SCROLL);
		scrollPosition = 0;
		repaint();
	}

	scrollLimit = 0;
	scroll = scrolling;
	if (scrolling != 0)
		scrollBackup = scrolling;
}

void Label::setScrollLimit(int limit)
{
	if (limit >= 0)
	{
		setScrolling(scrollBackup);
		scrollLimit = limit + 1;
	}
	else
	{
		setScrolling(0);
	}
}

void Label::update()
{
	string oldText = text;

	//Tooltips
	if ( !originalTooltip.empty() && TooltipHints != 0 )
	{
		strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());	

		if (dynamicTooltip)
			updateHint(hWnd, tooltip);
	}

	//Process Escape Sequence for ';'
	string process = originalText[current];
	while ( strstr(process.c_str(), "%~") != NULL )
	{
		process.replace(process.find("%~"), 2, ";", 1);
	}

	// Normal Text Processing
	if ( !noEscape )
	{
		text = systemInfo->processLabelText(process, this, &dynamicText);
	}
	// Skip Text Processing
	else
	{
		text = process;
		dynamicText = false;
	}

	if ( dynamicText || dynamicTooltip || alwaysUpdateContent )
	{
		SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
	}
	else
	{
		KillTimer(hWnd, TIMER_UPDATE);
	}

	if ( autoWidthMode != 0 || autoHeightMode != 0 )
		autoSize();

	if (stricmp(text.c_str(), oldText.c_str()) != NULL)
	{
		if(!textChangeCommand.empty())
			xLSExecute(hWnd, textChangeCommand.c_str(), SW_SHOWNORMAL);
	
		repaint();
	}
}

//------------------------------------------------------------
//Set Animations
//------------------------------------------------------------

void Label::setAnimation(bool start, int loops, int startFrame)
{
	imageLoop = loops;
	normalAnim = start;

	if (start)
	{
		frameCounter = startFrame;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}
	else
	{
		frameCounter = 0;
		loopCounter = 0;
		KillTimer(hWnd, TIMER_FRAMEANIMATION);
		repaint(true);
	}	
}

//------------------------------------------------------------
//Move/Resize/Reposition
//------------------------------------------------------------
//Type == 1 -> Move
//Type == 2 -> Resize
//Type == 3 -> Reposition

void Label::reposition(int type, int newx, int newy, int newwidth, int newheight, int steps, int time, bool repositionBy, bool movecommand, Label* sender)
{
	//Prevents infinite looping of docked moving
	this->sender = sender;

	bool moveAway = (0x8000 & GetKeyState(VK_SPACE));

	RECT rectPosition;
	if ( box )
	{
		rectPosition.left = actualx;
		rectPosition.right = actualx+actualwidth;
		rectPosition.top = actualy;
		rectPosition.bottom = actualy+actualheight;
	}
	else
	{
		GetWindowRect(hWnd, &rectPosition);
	}

	POINT posHolder;
	posHolder.x = rectPosition.left;
	posHolder.y = rectPosition.top;

	POINT sizeHolder;
	sizeHolder.x = (rectPosition.right-rectPosition.left);
	sizeHolder.y = (rectPosition.bottom-rectPosition.top);

	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		int counter = 0;
		int deltax = newx - rectPosition.left;
		int deltay = newy - rectPosition.top;
		int deltasizex = newwidth - (rectPosition.right-rectPosition.left);
		int deltasizey = newheight - (rectPosition.bottom-rectPosition.top);

		bool growx;
		bool growy;
		bool growsizex;
		bool growsizey;
		
		if (deltax < 0)
			growx = false;
		else
			growx = true;
		if (deltay < 0)
			growy = false;
		else
			growy = true;
		
		if (deltasizex < 0)
			growsizex = false;
		else
			growsizex = true;
		if (deltasizey < 0)
			growsizey = false;
		else
			growsizey = true;

		int restx = 0;
		int resty = 0;
		int restw = 0;
		int resth = 0;

		div_t div_result;
		div_result = div( deltax, steps );
		deltax = div_result.quot;
		restx = div_result.rem;
		div_result = div( deltay, steps );
		deltay = div_result.quot;
		resty = div_result.rem;

		div_result = div( deltasizex, steps );
		deltasizex = div_result.quot;
		restw = div_result.rem;
		div_result = div( deltasizey, steps );
		deltasizey = div_result.quot;
		resth = div_result.rem;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;
		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		if (deltasizex == 0 && growsizex)
			deltasizex = 1;
		else if (deltasizex == 0 && !growsizex)
			deltasizex = -1;
		if (deltasizey == 0 && growsizey)
			deltasizey = 1;
		else if (deltasizey == 0 && !growsizey)
			deltasizey = -1;

		int dynx = rectPosition.left;
		int dyny = rectPosition.top;
		int dynwidth = (rectPosition.right-rectPosition.left);
		int dynheight = (rectPosition.bottom-rectPosition.top);

		int stopper = 0;
		while (counter <= steps)
		{
			stopper = 0; // Stop if x, y, width and height are finished, stopper = 4
			if ( restx > 0 && deltax > 0 )
			{
				dynx = dynx+deltax+1;
				restx--;
			}
			else if ( restx < 0 && deltax < 0 )
			{
				dynx = dynx+deltax-1;
				restx++;
			}
			else
				dynx = dynx+deltax;

			if (growx && dynx > newx)
			{
				dynx = newx;
				stopper++;
			}
			else if (!growx && dynx < newx)
			{
				dynx = newx;
				stopper++;
			}

			if ( resty > 0 && deltay > 0 )
			{
				dyny = dyny+deltay+1;
				resty--;
			}
			else if ( resty < 0 && deltay < 0 )
			{
				dyny = dyny+deltay-1;
				resty++;
			}
			else
				dyny = dyny+deltay;

			if (growy && dyny > newy)
			{
				dyny = newy;
				stopper++;
			}
			else if (!growy && dyny < newy)
			{
				dyny = newy;
				stopper++;
			}

			if ( restw > 0 && deltasizex > 0 )
			{
				dynwidth = dynwidth+deltasizex+1;
				restw--;
			}
			else if ( restw < 0 && deltasizex < 0 )
			{
				dynwidth = dynwidth+deltasizex-1;
				restw++;
			}
			else
				dynwidth = dynwidth+deltasizex;

			if (growsizex && dynwidth > newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}
			else if (!growsizex && dynwidth < newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}

			if ( resth > 0 && deltasizey > 0 )
			{
				dynheight = dynheight+deltasizey+1;
				resth--;
			}
			else if ( resth < 0 && deltasizey < 0 )
			{
				dynheight = dynheight+deltasizey-1;
				resth++;
			}
			else
				dynheight = dynheight+deltasizey;

			if (growsizey && dynheight > newheight)
			{
				dynheight = newheight;
				stopper++;
			}
			else if (!growsizey && dynheight < newheight)
			{
				dynheight = newheight;
				stopper++;
			}
			
			SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
					SWP_NOACTIVATE | (type == 2 ? SWP_NOMOVE : 0) | (type == 1 ? SWP_NOSIZE : 0 ) | SWP_NOZORDER);
				
			if (box != NULL)
				SendMessage(box, WM_PAINT, 0, 0);

			SendMessage(hWnd, WM_PAINT, 0, 0);

			if (stopper == 4)
				counter = steps;

			counter++;

			Sleep(time);
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
						SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
		}
		else if ( type == 3 )
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOZORDER);
		}	
	}
	else
	{
		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
						SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
		}
		else if ( type == 3 )
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOZORDER);
		}	
	}

	SendMessage(hWnd, WM_PAINT, 0, 0);

	if(type != 2 && movecommand && labelInitialized && !moveCommand.empty() && (newx != posHolder.x || newy != posHolder.y))
		xLSExecute(hWnd, moveCommand.c_str(), SW_SHOWNORMAL);
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void Label::show(bool execute, bool textEditBox)
{
	if ( textEditBox )
	{
		if ( hWndTextBox )
		{
			ShowWindow(hWndTextBox, SW_SHOW);
			textboxVisible = true;
			SetFocus(hWndTextBox);
		}
	}
	else
	{
		KillTimer(hWnd, TIMER_AUTOHIDE);

		if (alphaFade && !visible)
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);

			KillTimer(hWnd, TIMER_ALPHAFADEOUT);
			SetTimer(hWnd, TIMER_ALPHAFADEIN, alphaFadeDelay, NULL);
		}
		else
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);

		if(!showCommand.empty() && execute && !visible)
			xLSExecute(hWnd, showCommand.c_str(), SW_SHOWNORMAL);

		visible = true;

		if (normalAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
		}

		repaint(true);

		if ( box != NULL && parentIsAlphaMapped )
			SendMessage(box, LM_EXTERNALDRAWNOTIFY, 0, 0 );

		//Call Initial Zorder
		if ( !afterClass.empty() )
			SetWindowPos(hWnd, NULL, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
	}
}

void Label::showHide(double timeout, bool execute)
{
	if ( !visible )
		show(execute);
	else
		hide(execute);

	SetTimer(hWnd, TIMER_AUTOHIDE, timeout * 1000, NULL);
}

void Label::hide(bool execute, bool textEditBox)
{
	if ( textEditBox )
	{
		if ( hWndTextBox )
		{
			SetWindowText(hWndTextBox,"");
			ShowWindow(hWndTextBox, SW_HIDE);
			textboxVisible = false;
		}
	}
	else
	{
		if (alphaFade && visible)
		{
			UpdateAlpha = alphaTransparency;

			KillTimer(hWnd, TIMER_ALPHAFADEIN);
			SetTimer(hWnd, TIMER_ALPHAFADEOUT, alphaFadeDelay, NULL);
		}
		else
			ShowWindow(hWnd, SW_HIDE);

		if(!hideCommand.empty() && execute && visible)
			xLSExecute(hWnd, hideCommand.c_str(), SW_SHOWNORMAL);

		visible = false;

		KillTimer(hWnd, TIMER_AUTOHIDE);

		if (normalAnim || hoverAnim)
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

		if ( box != NULL && parentIsAlphaMapped )
			SendMessage(box, LM_EXTERNALDRAWNOTIFY, 0, 0 );
	}
}


//------------------------------------------------------------
//Clipboard Functions
//------------------------------------------------------------

void Label::clipboardCopy(string prefix)
{
	string toCopy = prefix.empty() ? text : (prefix + " " + text);
	SetClipboardText(hWnd, toCopy.c_str());
}

void Label::clipboardPaste(string prefix)
{
	char buffer[256];

	if(GetClipboardText(hWnd, buffer, 256))
	{
		string toPaste = prefix.empty() ? string(buffer) : (prefix + " " + string(buffer));
		setText(toPaste);
	}
}

//------------------------------------------------------------
//Region Events (Checking)
//------------------------------------------------------------

bool Label::checkEnterLeaveRegion(int x, int y, StringList inputlist)
{
	bool found = false;
	if (!inputlist.empty())
	{	
		char left[16], top[16], right[16], bottom[16], command[MAX_PATH_LENGTH];
		char *buffers[] = {left, top, right, bottom};

		int intleft = 0;
		int intright = 0;
		int inttop = 0;
		int intbottom = 0;

		for(StringListIterator sit = inputlist.begin(); sit != inputlist.end(); sit++)
		{
			if(LCTokenize((*sit).c_str(), buffers, 4, command) >= 4)
			{
				intleft = atoi(left);
				intright = atoi(right);
				inttop = atoi(top);
				intbottom  = atoi(bottom);

				if( x >= intleft && x <= intright && y >= inttop && y <= intbottom )
				{
					if (!handcursor)
					{
						if ( !IsOS(OS_2KXP) )
						{
							// Get the windows directory
							CString strWndDir;
							GetWindowsDirectory(strWndDir.GetBuffer(MAX_PATH_LENGTH), MAX_PATH_LENGTH);
							strWndDir.ReleaseBuffer();

							strWndDir += _T("\\winhlp32.exe");

							// This retrieves cursor #106 from winhlp32.exe, which is a hand pointer
							HMODULE hModule = LoadLibrary(strWndDir);
								
							if (hModule)
								handcursor = ::LoadCursor(hModule, MAKEINTRESOURCE(106));
					
							FreeLibrary(hModule);
						}
						else
							handcursor = LoadCursor( 0, MAKEINTRESOURCE(32649) );
					}

					if (handcursor)
						SetCursor(handcursor);
						
					//Processing Hover Change
					if (hrefconfig.size() == 2 && !mouseInsideHREF)
					{
						CString textprocess = originalText[current].c_str();
						CString textoutput = originalText[current].c_str();
						CString commandprocess = command;

						textprocess.MakeLower();
						commandprocess.MakeLower();

						int pos = 0;
						textprocess.Replace("&quot;", "\"");
						textprocess.Replace("&lt;", "<");
						textprocess.Replace("&gt;", ">");
						textprocess.Replace("&amp;", "&");
						//textprocess.Replace("%~", ";");
						//textprocess.Replace("[semicolon]", ";"); 
						textoutput.Replace("&quot;", "\"");
						textoutput.Replace("&lt;", "<");
						textoutput.Replace("&gt;", ">");
						textoutput.Replace("&amp;", "&");
						//textoutput.Replace("%~", ";");
						//textoutput.Replace("[semicolon]", ";"); 

						pos = textprocess.Find(commandprocess);
						if ( pos != -1 )
						{
							pos = textprocess.Find("'>", pos);
							if ( pos != -1 )
							{
								textprocess.Insert(pos+2, hrefconfig[0].c_str());
								textoutput.Insert(pos+2, hrefconfig[0].c_str());
								pos = textprocess.Find("</a>", pos+2);
								if ( pos != -1 )
								{
									textprocess.Insert(pos, hrefconfig[1].c_str());
									textoutput.Insert(pos, hrefconfig[1].c_str());

									textoutput.Replace("\"", "");

									setText((string)textoutput);
									
									mouseInsideHREF = true;
								}
							}
						}
					}
					//End

					found = true;
					break;
				}
			}
		}
	}

	return found;
}

bool Label::checkRegion(int x, int y, StringList inputlist, bool internal)
{
	bool found = false;
	if (!inputlist.empty())
	{	
		char left[16], top[16], right[16], bottom[16], command[MAX_PATH_LENGTH];
		char *buffers[] = {left, top, right, bottom};

		int intleft = 0;
		int intright = 0;
		int inttop = 0;
		int intbottom = 0;

		for(StringListIterator sit = inputlist.begin(); sit != inputlist.end(); sit++)
		{
			if(LCTokenize((*sit).c_str(), buffers, 4, command) >= 4)
			{
				if ( !internal )
				{
					intleft = xParseCoord(left, 0, actualwidth);
					intright = xParseCoord(right, actualwidth, actualwidth);
					inttop = xParseCoord(top, 0, actualheight);
					intbottom  = xParseCoord(bottom, actualheight, actualheight);
				}
				else
				{
					intleft = atoi(left);
					intright = atoi(right);
					inttop = atoi(top);
					intbottom  = atoi(bottom);
				}

				if( x >= intleft && x <= intright && y >= inttop && y <= intbottom )
				{	
					if ( internal )
					{
						CString process = command;
						strcpy(command, process.Left(process.Find("'>")));
					}
				
					xLSExecute(hWnd, command, SW_SHOWNORMAL);

					found = true;
					break;	
				}
			}
		}
	}
	return found;
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

LRESULT Label::NcHitTest(int x, int y)
{
	if ( y >= 0 && y <= resizeBorderTop && x >= 0 && x <= resizeBorderLeft )
	{
		return HTTOPLEFT;
	}
	else if ( y >= 0 && y <= resizeBorderTop && x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{	
		return HTTOPRIGHT;
	}	
	else if ( y >= 0 && y <= resizeBorderTop )
	{		
		return HTTOP;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight && x >= 0 && x <= resizeBorderLeft )
	{	
		return HTBOTTOMLEFT;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight && x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{		
		return HTBOTTOMRIGHT;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight )
	{	
		return HTBOTTOM;
	}	
	else if ( x >= 0 && x <= resizeBorderLeft )
	{	
		return HTLEFT;
	}	
	else if ( x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{	
		return HTRIGHT;
	}	
	else
		return 0;
}

void Label::onLButtonDblClk(int x, int y)
{
	if(!leftDoubleClickCommand.empty())
		xLSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onLButtonDown(int x, int y)
{
	// 1 = LeftButton
	ontemplateButtonDown(1, leftClickDownCommand);
}

void Label::onLButtonUp(int x, int y)
{	
	// 1 = LeftButton
	ontemplateButtonUp(x, y, 1,labelLeftClickRegions, leftClickCommand, leftClickUpCommand);
}

void Label::onMButtonDblClk(int x, int y)
{
	if(!middleDoubleClickCommand.empty())
		xLSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onMButtonDown(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonDown(3, middleClickDownCommand);
}

void Label::onMButtonUp(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonUp(x, y, 3, labelMiddleClickRegions, middleClickCommand, middleClickUpCommand);
}

void Label::onRButtonDblClk(int x, int y)
{
	if(!rightDoubleClickCommand.empty())
		xLSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onRButtonDown(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonDown(2, rightClickDownCommand);
}

void Label::onRButtonUp(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonUp(x, y, 2, labelRightClickRegions, rightClickCommand, rightClickUpCommand);
}

void Label::onWheelDown(int x, int y)
{
	if(!wheelDownCommand.empty())
		xLSExecute(hWnd, wheelDownCommand.c_str(), SW_SHOWNORMAL);

	if (focusOnEnter == 1)
	{
		focusOnEnter = 2;
		mouse_event(MOUSEEVENTF_MIDDLEDOWN, NULL, NULL, NULL, NULL);
		mouse_event(MOUSEEVENTF_MIDDLEUP, NULL, NULL, NULL, NULL);
	}
}

void Label::onWheelUp(int x, int y)
{
	if(!wheelUpCommand.empty())
		xLSExecute(hWnd, wheelUpCommand.c_str(), SW_SHOWNORMAL);

	if (focusOnEnter == 1)
	{
		focusOnEnter = 2;
		mouse_event(MOUSEEVENTF_MIDDLEDOWN, NULL, NULL, NULL, NULL);
		mouse_event(MOUSEEVENTF_MIDDLEUP, NULL, NULL, NULL, NULL);
	}
}

void Label::onMouseEnter()
{
	SendMessage(hWnd, LM_SETCURSOR, 0, 0);

	if (useHover)
	{
		//Kill AnimTimer from Normal Animation
		if (normalAnim)
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

		//If HoverAnim is set (hoverframecount != 0)
		if (hoverAnim)
		{
			hoverActive = true;
			pressedActive = false;
			frameCounter = 1;
			loopCounter = 0;
			//Start AnimTimer from Hover Animation
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
		//If HoverAnim is not set (hoverframecount == 0)
		else
		{
			frameCounter = 0;
			hoverActive = true;
			pressedActive = false;
			repaint(true);
		}
	}
	
	if(!enterCommand.empty())
		xLSExecute(hWnd, enterCommand.c_str(), SW_SHOWNORMAL);
		
	if (focusOnEnter == 1)
	{
		focusOnEnter = 2;
		mouse_event(MOUSEEVENTF_MIDDLEDOWN, NULL, NULL, NULL, NULL);
		mouse_event(MOUSEEVENTF_MIDDLEUP, NULL, NULL, NULL, NULL);
	}
}

void Label::onMouseLeave()
{
	if (useHover || usePressed)
	{
		//If HoverAnim was active (hoverframecount != 0)
		if (hoverAnim)
		{
			hoverActive = false;
			pressedActive = false;
			frameCounter = 0;
			//Kill AnimTimer from Hover Animation
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				loopCounter = 0;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
				repaint(true);
		}
		else
		{
			hoverActive = false;
			pressedActive = false;
			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
				repaint(true);
		}
	}

	if(!leaveCommand.empty())
		xLSExecute(hWnd, leaveCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onMouseMove(int x, int y)
{
	if(moveable)
	{
		if (mousePressed && movemodifierkeyPressed && moveButtonPressed == moveButton)
		{
			KillTimer(hWnd, TIMER_MOUSETRACK);

			POINT newpt = {x,y};
			ClientToScreen(hWnd, &newpt);
			if ( box )
				ScreenToClient(box, &newpt);
			newpt.x = newpt.x-savedpt.x;
			newpt.y = newpt.y-savedpt.y;

			if ( !hasMoved && (newpt.x != actualx || newpt.y != actualy) )
				hasMoved = true;

			if ( fireOnMoveDuringMove )
				reposition(1, newpt.x, newpt.y, 0, 0);
			else
				reposition(1, newpt.x, newpt.y, 0, 0, 0, 0, false, false);

			if ( hasMoved )
				SendMessage(hWnd, LM_SETCURSOR, 0, (long) IDC_SIZEALL);
		}
		else
		{
			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);

			if(!mouseInside)
			{
				mouseInside = true;
				onMouseEnter();
			}

			if (!labelHREFClickRegions.empty())
			{
				if (!checkEnterLeaveRegion(x, y, labelHREFClickRegions) && mouseInsideHREF)
				{
					if (!originalTextBackup.empty())
					{
						current = currentBackup;
						setText(originalTextBackup);
					}
					SendMessage(hWnd, LM_SETCURSOR, 0, 0);
					mouseInsideHREF = false;
				}
			}
		}
	}
	else if (!moveable)
	{
		if(!mouseInside)
		{
			mouseInside = true;
			onMouseEnter();
			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
		}

		if (!labelHREFClickRegions.empty())
		{
			if (!checkEnterLeaveRegion(x, y, labelHREFClickRegions) && mouseInsideHREF)
			{
				if (!originalTextBackup.empty())
				{
					current = currentBackup;
					setText(originalTextBackup);
				}
				SendMessage(hWnd, LM_SETCURSOR, 0, 0);
				mouseInsideHREF = false;
			}
		}
	}
}

//------------------------------------------------------------
//AutoWidthMode and AutoHeightMode Implementation
//------------------------------------------------------------
void Label::autoSize()
{
	if ( !backgroundDC )
		return;

	RECT rectPosition;
	if ( box )
	{
		rectPosition.left = actualx;
		rectPosition.right = actualx+actualwidth;
		rectPosition.top = actualy;
		rectPosition.bottom = actualy+actualheight;
	}
	else
	{
		GetWindowRect(hWnd, &rectPosition);
	}

	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	if ( width <= 0 || height <= 0 )
		return;

	int textWidth = 0;
	int textHeight = 0;

	POINT size = font->measure(backgroundDC, leftBorder, width - rightBorder, text, align, vertAlign);

	textWidth = size.x+6;
	textHeight = size.y+6;

	if ( oldWidth == textWidth && oldHeight == textHeight )
	{
		return;
	}
	else
	{
		oldWidth = textWidth;
		oldHeight = textHeight;
	}

	//For AUTOWIDTH/HEIGHT MODE	
	int newx;
	int newy;
	int newwidth;
	int newheight;

	int deltax;
	int deltay;

	if (autoWidthMode != 0)
		deltax = textWidth - (width - leftBorder - rightBorder);
	else
		deltax = 0;

	if (autoHeightMode != 0)
		deltay = textHeight - (height - topBorder - bottomBorder);
	else
		deltay = 0;

	newx = rectPosition.left;
	newy = rectPosition.top;
	newwidth = width + deltax;
	newheight = height + deltay;

	//Check MIN/MAX
	if (newwidth < autoMinWidth || newwidth > autoMaxWidth)
	{
		if (newwidth < autoMinWidth)
		{
			deltax = deltax + (autoMinWidth-newwidth);
			newwidth = autoMinWidth;
		}
		if (newwidth > autoMaxWidth)
		{
			deltax = deltax - (newwidth-autoMaxWidth);
			newwidth = autoMaxWidth;
		}
	}
	if (newheight < autoMinHeight || newheight > autoMaxHeight)
	{
		if (newheight < autoMinHeight)
		{
			deltay = deltay + (autoMinHeight-newheight);
			newheight = autoMinHeight;
		}
		if (newheight > autoMaxHeight)
		{
			deltay = deltay - (newheight-autoMaxHeight);
			newheight = autoMaxHeight;
		}
	}

	if ((autoWidthMode == 0 || autoWidthMode == 1) && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = rectPosition.left;
		newy = rectPosition.top;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 2)
	{
		newx = rectPosition.left;
		newy = rectPosition.top-deltay;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 3)
	{
		newx = rectPosition.left;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 2 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = rectPosition.left-deltax;
		newy = rectPosition.top;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 2)
	{
		newx = rectPosition.left-deltax;
		newy = rectPosition.top-deltay;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 3)
	{
		newx = rectPosition.left-deltax;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 3 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = startCenterX-(newwidth/2);
		newy = rectPosition.top;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 2)
	{
		newx = startCenterX-(newwidth/2);
		newy = rectPosition.top-deltay;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 3)
	{
		newx = startCenterX-(newwidth/2);
		newy = startCenterY-(newheight/2);
	}

	reposition(3, newx, newy, newwidth, newheight);
}

void Label::repaint(bool invalidateCache)
{
	if(invalidateCache)
	{
		if(backgroundDC && backgroundBitmap)
		{
			backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
			DeleteObject(backgroundBitmap);
			backgroundBitmap = 0;
		}

		if(backupDC && backupBitmap)
		{
			backupBitmap = (HBITMAP) SelectObject(backupDC, backupBitmap);
			DeleteObject(backupBitmap);
			backupBitmap = 0;
		}
	}

	if ( !alphaMap )
		InvalidateRect(hWnd, NULL, false);

	if ( hWndTextBox )
		InvalidateRect(hWndTextBox, NULL, false);

	if (transparencyMode == 1 && labelInitialized)
		updateRegion();

	//Attention!! Background MUST exist
	if ( alphaMap && IsOS(OS_2KXP) && background )
		UpdateLabelAlpha(alphaTransparency);

	if ( box != NULL && parentIsAlphaMapped )
		SendMessage(box, LM_EXTERNALDRAWNOTIFY, 0, 0 );
}

void Label::onPaint(HDC hDC)
{
	bool useFrames = false;

	RECT rectPosition;
	if ( box )
	{
		rectPosition.left = actualx;
		rectPosition.right = actualx+actualwidth;
		rectPosition.top = actualy;
		rectPosition.bottom = actualy+actualheight;
	}
	else
	{
		GetWindowRect(hWnd, &rectPosition);
	}

	int x = rectPosition.left;
	int y = rectPosition.top;
	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	if ( width <= 0 || height <= 0 )
		return;

	rectPosition.left = 0;
	rectPosition.top = 0;
	rectPosition.right = width;
	rectPosition.bottom = height;

	// keep a cached rendition of the background
	if(!backgroundBitmap)
	{
		if(!backgroundDC)
			backgroundDC = CreateCompatibleDC(hDC);

		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		SelectObject(backgroundDC, backgroundBitmap);

		//True Transparency
		if (transparencyMode == 1 || box)
		{
			::SetBkColor(backgroundDC, RGB(255, 0, 255));
			::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);

			if ( transparencyMode == 2 )
				transparencyMode = 1;
		}
		//PrePaint Background (Fake Transparency)
		else if(transparencyMode == 2)
		{
			// paint desktop on display DC and then into background buffer
			PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, FALSE);
		}
		//Default Background
		else if (transparencyMode == 3)
		{
			if (borderMethod == 2)
				DrawEdge(backgroundDC, &rectPosition, EDGE_SUNKEN, BF_RECT | BF_SOFT | BF_MIDDLE);
			else if (borderMethod == 3)
				DrawEdge(backgroundDC, &rectPosition, EDGE_ETCHED, BF_RECT | BF_SOFT | BF_MIDDLE);
			else if (borderMethod == 4)
				DrawEdge(backgroundDC, &rectPosition, EDGE_BUMP, BF_RECT | BF_SOFT | BF_MIDDLE);
			else
				DrawEdge(backgroundDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT | BF_MIDDLE);
		}
		
		//Normal Background Painting if no Animation
		if ( (!normalAnim && !hoverAnim) || frameCounter == 0 || (usePressed && pressedActive))
		{
			if (hoverActive && useHover)
			{
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 1);
			}
			else if (pressedActive && usePressed)
			{
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 2);
			}
			else
			{
				background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 0);
			}
		}
		//Special Background Painting if Animation (normal or hover)
		else
		{
			useFrames = true;
			if (hoverActive && useHover)
				background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 1, frameCounter);
			else if (pressedActive && usePressed)
				background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 1, frameCounter);
			else
				background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 0, frameCounter);
		}

		if ( !ghostlabelList.empty() )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				if ( (hoverActive && useHover && (*glit)->hover) || (pressedActive && usePressed && (*glit)->pressed) || (!hoverActive && !pressedActive && (*glit)->normal) )
				{
					if ( (*glit)->alphamap || (*glit)->alpha != 255 )
					{
						BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, (*glit)->alpha, AC_SRC_ALPHA };

						AlphaBlend(backgroundDC, (*glit)->x, (*glit)->y, ((*glit)->fullstretch ? width : (*glit)->width), ((*glit)->fullstretch ? height : (*glit)->height), (*glit)->memoryDC, 0, 0, (*glit)->width, (*glit)->height, blendFunction);
					}
					else
					{
						TransparentBltLS(backgroundDC, (*glit)->x, (*glit)->y, ((*glit)->fullstretch ? width : (*glit)->width), ((*glit)->fullstretch ? height : (*glit)->height), (*glit)->memoryDC, 0, 0, RGB(255, 0, 255));
					}
				}
			}
		}
	}

	//Only needed with Text, AlphaMapHooking or BorderMethod
	if ( !text.empty() || 
		(hasHookedModules && (alphaMap || parentIsAlphaMapped)) || 
		(borderMethod != 0 || hoverborderMethod != 0 || pressedborderMethod != 0) )
	{
		// double buffer for flicker-free paint
		HDC	bufferDC = CreateCompatibleDC(hDC);

		HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
		
		SelectObject(bufferDC, bufferBitmap);

		// blt background into double buffer
		BitBlt(bufferDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

		//Text Painting Function
		if ( !text.empty() && !useTextFade )
			onTextPaint(bufferDC, width, height);
		else if ( !text.empty() && useTextFade )
			TextFadeProcessing(bufferDC, width, height);

		//Hooked Child Window AlphaMap Painting
		if ( hasHookedModules && (alphaMap || parentIsAlphaMapped) )
		{
			for(HWND hWndChild = ::GetTopWindow(hWnd); hWndChild != NULL; hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT))
			{
				SendMessage(hWndChild, LM_EXTERNALDRAW, (long) bufferDC, 0 );
			}
		}

		//Border Painting
		if (hoverActive && useHover && hoverborderMethod != 0)
		{
			if (hoverborderMethod == 1)
				DrawEdge(bufferDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT);
			else if (hoverborderMethod == 2)
				DrawEdge(bufferDC, &rectPosition, EDGE_SUNKEN, BF_RECT | BF_SOFT);
			else if (hoverborderMethod == 3)
				DrawEdge(bufferDC, &rectPosition, EDGE_ETCHED, BF_RECT | BF_SOFT);
			else if (hoverborderMethod == 4)
				DrawEdge(bufferDC, &rectPosition, EDGE_BUMP, BF_RECT | BF_SOFT);
		}
		else if (pressedActive && usePressed && pressedborderMethod != 0)
		{
			if (pressedborderMethod == 1)
				DrawEdge(bufferDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT);
			else if (pressedborderMethod == 2)
				DrawEdge(bufferDC, &rectPosition, EDGE_SUNKEN, BF_RECT | BF_SOFT);
			else if (pressedborderMethod == 3)
				DrawEdge(bufferDC, &rectPosition, EDGE_ETCHED, BF_RECT | BF_SOFT);
			else if (pressedborderMethod == 4)
				DrawEdge(bufferDC, &rectPosition, EDGE_BUMP, BF_RECT | BF_SOFT);
		}
		else if (borderMethod != 0)
		{
			if (borderMethod == 1)
				DrawEdge(bufferDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT);
			else if (borderMethod == 2)
				DrawEdge(bufferDC, &rectPosition, EDGE_SUNKEN, BF_RECT | BF_SOFT);
			else if (borderMethod == 3)
				DrawEdge(bufferDC, &rectPosition, EDGE_ETCHED, BF_RECT | BF_SOFT);
			else if (borderMethod == 4)
				DrawEdge(bufferDC, &rectPosition, EDGE_BUMP, BF_RECT | BF_SOFT);
		}

		// blt the double buffer to the display
		BitBlt(hDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);

		if ( parentIsAlphaMapped )
		{
			if(!backupBitmap)
			{
				if(!backupDC)
					backupDC = CreateCompatibleDC(hDC);

				backupBitmap = CreateCompatibleBitmap(hDC, width, height);
				SelectObject(backupDC, backupBitmap);
			}

			BitBlt(backupDC, 0, 0, width, height, bufferDC, 0, 0, SRCCOPY);
		}

		if(bufferBitmap)
		{
			bufferBitmap = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
			DeleteObject(bufferBitmap);
		}
		if (bufferDC)
			DeleteDC(bufferDC);
	}
	else
	{
		// blt the background buffer to the display
		BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

		if ( parentIsAlphaMapped )
		{
			if(!backupBitmap)
			{
				if(!backupDC)
					backupDC = CreateCompatibleDC(hDC);

				backupBitmap = CreateCompatibleBitmap(hDC, width, height);
				SelectObject(backupDC, backupBitmap);
			}

			BitBlt(backupDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
		}
	}
}

void Label::onTextPaint(HDC desthdc, int width, int height)
{
	//--------------------------------------------------------------------------------
	// start rendering text
	//--------------------------------------------------------------------------------
	//Create Region in which text drawing is allowed
	HRGN drawArea = CreateRectRgn( 0, 0, width, height );
	HRGN allowedArea = CreateRectRgn( leftBorder, topBorder, width - rightBorder, height - bottomBorder );
	SelectClipRgn(desthdc, allowedArea);
	DeleteObject(allowedArea);
	
	int textWidth;
	int textHeight;

	POINT size = font->measure(desthdc, leftBorder, width - rightBorder, text, align, vertAlign);

	textWidth = size.x+6;
	textHeight = size.y+6;

	if ( ((scroll == 1 || scroll == 3) && textWidth > width - leftBorder - rightBorder) || ((scroll == 2 || scroll == 4) && textHeight > height - topBorder - bottomBorder) )
	{			
		if (scroll == 1 || scroll == 3)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = textWidth + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > textWidth + scrollPadLength)
				scrollPosition = 0;
		}
		else if (scroll == 2 || scroll == 4)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = textHeight + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > textHeight + scrollPadLength )
				scrollPosition = 0;
		}
		
		if (scroll == 1 || scroll == 3)
		{
			font->apply(desthdc,
						scrollPosition,
						scrollPosition - scrollPadLength,
						topBorder,
						height - bottomBorder,
						text,
						0,
						vertAlign,
						labelHREFClickRegions);

			StringList tmplabelHREFClickRegions;
			tmplabelHREFClickRegions.clear();

			font->apply(desthdc,
						scrollPosition - scrollPadLength - textWidth,
						scrollPosition + textWidth,
						topBorder,
						height - bottomBorder,
						text,
						0,
						vertAlign,
						tmplabelHREFClickRegions);

			labelHREFClickRegions.merge(tmplabelHREFClickRegions);
			tmplabelHREFClickRegions.clear();
		}
		else if (scroll == 2 || scroll == 4)
		{
			font->apply(desthdc,
					leftBorder,
					width - rightBorder,
					scrollPosition,
					scrollPosition - scrollPadLength,
					text,
					align,
					2,
					labelHREFClickRegions);

			StringList tmplabelHREFClickRegions;
			tmplabelHREFClickRegions.clear();

			font->apply(desthdc,
					leftBorder,
					width - rightBorder,
					scrollPosition - scrollPadLength - textHeight,
					scrollPosition + textHeight,
					text,
					align,
					2,
					tmplabelHREFClickRegions);

			labelHREFClickRegions.merge(tmplabelHREFClickRegions);
			tmplabelHREFClickRegions.clear();
		}
	}
	else
	{
		font->apply(desthdc,
					leftBorder,
					width - rightBorder,
					topBorder,
					height - bottomBorder,
					text,
					align,
					vertAlign,
					labelHREFClickRegions);
	}

	//Reset Region in which text drawing is allowed
	SelectClipRgn(desthdc, drawArea);
	DeleteObject(drawArea);
	//--------------------------------------------------------------------------------
	// stop rendering text
	//--------------------------------------------------------------------------------
}

void Label::updateRegion()
{	
	RECT rectPosition;
	GetWindowRect(hWnd, &rectPosition);

	if ( !alphaMap && transparencyNeeded )
	{
		HDC tmpDC = CreateCompatibleDC(NULL);
		HBITMAP tmpBmp = CreateBitmap((rectPosition.right-rectPosition.left), (rectPosition.bottom-rectPosition.top), 1, bitsperpixel, NULL);
		HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

		//Draw Image
		onPaint(tmpDC);
	
		SelectObject(tmpDC, tmpObj);

		HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
		HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
		CombineRgn(windowRgn, region, NULL, RGN_COPY);

		SetWindowRgn(hWnd, windowRgn, true);	

		DeleteObject(windowRgn);
		DeleteObject(region);
		DeleteObject(tmpBmp);
		DeleteObject(tmpObj);
		DeleteDC(tmpDC);
	}
	else if ( alphaMap )
	{
		HRGN windowRgn = CreateRectRgn(0, 0, (rectPosition.right-rectPosition.left), (rectPosition.bottom-rectPosition.top));
		SetWindowRgn(hWnd, windowRgn, false);
		DeleteObject(windowRgn);
	}
}

bool Label::FindDockedLabel(Label* findLabel)
{
	for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
	{
		if ( (*lit) == findLabel )
			return true;
	}
	return false;
}

void Label::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if ( (windowPos->flags & SWP_NOZORDER) == 0 )
	{
		if ( !afterClass.empty() )
		{
			HWND before = NULL;
			if ( !afterClass.empty() )
			{
				before = FindWindowEx2(afterClass, afterTitle);
		
				if ( before != NULL )
					windowPos->hwndInsertAfter = before;
			}
			else if ( !alwaysOnTop )
				windowPos->flags = windowPos->flags | SWP_NOZORDER;
		}
		else if ( !alwaysOnTop )
			windowPos->flags = windowPos->flags | SWP_NOZORDER;
	}

	if ( (windowPos->flags & SWP_NOSIZE) == 0 && alphaMap && IsOS(OS_2KXP) && background )
		repaint(true);

	if ( (windowPos->flags & SWP_NOMOVE) != 0 )
		return;

	int result = 0;

	//Moving of hooked Label out of the Parent isn't allowed
	if ( mousePressed && movemodifierkeyPressed && box )
	{
		RECT boxPosition;
		GetWindowRect(box, &boxPosition);
		int maxX = (boxPosition.right-boxPosition.left);
		int maxY = (boxPosition.bottom-boxPosition.top);

		// left edge
		if ( windowPos->x < 0 )
		{
			windowPos->x = 0;
			result = 1;
		}

		// top edge
		if ( windowPos->y < 0 )
		{
			windowPos->y = 0;
			result = 1;
		}

		// bottom edge
		if ( windowPos->y+actualheight > maxY)
		{
			windowPos->y = maxY-actualheight;
			result = 1;
		}

		// right edge
		if ( windowPos->x+actualwidth > maxX)
		{
			windowPos->x = maxX-actualwidth;
			result = 1;
		}

		if ( result != 0 )
			return;
	}

	//Check Snap to Borders
	if ( mousePressed && movemodifierkeyPressed && snapDistance != 0 )
	{
		POINT pos = {windowPos->x, windowPos->y};
		POINT size = {actualwidth, actualheight};

		//Check First WorkArea Border
		if ( (stricmp("alllabels", snapTo.c_str()) == NULL || stricmp("desktop", snapTo.c_str()) == NULL) )
		{
			RECT rScreen;
			SystemParametersInfo(SPI_GETWORKAREA, 0, &rScreen, NULL);
			result = isCloseTo( pos, size, rScreen.left, rScreen.top, rScreen.right-rScreen.left, rScreen.bottom-rScreen.top, snapDistance);
		}

		string Bang = "";
		if (result == 0)
		{
			for( LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( (*lit)->getBox() == NULL && (*lit)->isVisible() && stricmp(name.c_str(), (*lit)->getName().c_str()) != NULL)
				{
					if ( stricmp((*lit)->getName().c_str(), snapTo.c_str()) == NULL )
					{
						result = isCloseTo( pos, size, (*lit)->actualx, (*lit)->actualy, (*lit)->actualwidth, (*lit)->actualheight, snapDistance);
						if ( result == 0 )
						{
							if ( dockOnSnapTo == 1 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 2 )
							{
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 3 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);	
								}
							}
						}
					}
					if ( stricmp((*lit)->labelGroup.c_str(), snapTo.c_str()) == NULL || stricmp("alllabels", snapTo.c_str()) == NULL )
					{
						result = isCloseTo( pos, size, (*lit)->actualx, (*lit)->actualy, (*lit)->actualwidth, (*lit)->actualheight, snapDistance);
						if ( result == 0 )
						{
							if ( dockOnSnapTo == 1 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 2 )
							{
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 3 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
						}
					}
					if (result != 0)
					{
						if ( dockOnSnapTo == 1 )
						{
							Bang = (name + " " + (*lit)->getName());
							snapToList.insert(snapToList.end(), Bang);
						}
						else if ( dockOnSnapTo == 2 )
						{
							Bang = ((*lit)->getName() + " " + name);
							snapToList.insert(snapToList.end(), Bang);
						}
						else if ( dockOnSnapTo == 3 )
						{
							Bang = (name + " " + (*lit)->getName());
							snapToList.insert(snapToList.end(), Bang);
							
							Bang = ((*lit)->getName() + " " + name);
							snapToList.insert(snapToList.end(), Bang);
						}

						break;
					}
				}
			}
		}
		
		if (result == 1)
		{
			windowPos->x = pos.x;
		}
		else if (result == 2)
		{	
			windowPos->y = pos.y;
		}
		else if (result == 3)
		{
			windowPos->x = pos.x;
			windowPos->y = pos.y;
		}
	}
}

void Label::onMove(int x, int y)
{
	//Docked Label Must Move (MoveBy)
	if ( !hasDockedLabelList.empty() )
	{
		for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
		{
			if ( (*lit) == this || (*lit) == sender )
				continue;
			
			(*lit)->reposition(1, x-oldPositionBackup.x, y-oldPositionBackup.y, 0, 0, 0, 0, true, true, this);
		}
	}

	oldPositionBackup.x = actualx = x;
	oldPositionBackup.y = actualy = y;

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = actualx+(actualwidth/2);
		startCenterY = actualy+(actualheight/2);
	}

	if ( labelInitialized )
	{
		PrefixedLSSetVariable(name, "CurrentX", IntToString( actualx ));
		PrefixedLSSetVariable(name, "CurrentY", IntToString( actualy ));
	}

	if ( transparencyMode == 2 && transparencyNeeded )
		repaint(true);
}

void Label::onSize(int width, int height)
{
	//Docked Label Must Move (MoveBy)
	if ( !hasDockedLabelList.empty() )
	{
		for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
		{
			if ( (*lit) == this || (*lit) == sender || (*lit)->dockedtoSide == 0 || (*lit)->dockedtoSide == 1 || (*lit)->dockedtoSide == 4 )
				continue;

			if ( (*lit)->dockedtoSide == 2 && width == actualwidth )
				continue;
			else if ( (*lit)->dockedtoSide == 3 && height == actualheight )
				continue;
			else if ( (*lit)->dockedtoSide == 5 && width == actualwidth )
				continue;
			else if ( (*lit)->dockedtoSide == 6 && height == actualheight )
				continue;
			else if ( (*lit)->dockedtoSide == 7 && width == actualwidth && height == actualheight)
				continue;
			
			(*lit)->reposition(1, width-oldSizeBackup.x, height-oldSizeBackup.y, 0, 0, 0, 0, true, true, this);
		}
	}

	oldSizeBackup.x = actualwidth = width;
	oldSizeBackup.y = actualheight = height;

	if ( alphaMap && IsOS(OS_2KXP) && background )
		UpdateLabelAlpha(alphaTransparency);
	else
		repaint(true);

	if (labelInitialized)
	{
		PrefixedLSSetVariable(name, "CurrentWidth", IntToString( width ));
		PrefixedLSSetVariable(name, "CurrentHeight", IntToString( height ));

		if ( !ghostlabelList.empty() )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				(*glit)->x = xParseCoord((*glit)->orgx, 0, actualwidth);
				(*glit)->y = xParseCoord((*glit)->orgy, 0, actualheight);
			}
		}
	}

	//Set new Rect for ToolTip Tracker
	if ( TooltipHints )
		updateHint(hWnd, tooltip, true);

	if ( hWndTextBox )
	{
		SetWindowPos(hWndTextBox, NULL, 
						xParseCoord(textboxXOrg, 0, actualwidth),
						xParseCoord(textboxYOrg, 0, actualheight),
						xParseDimen(textboxWidthOrg, actualwidth, actualwidth), 
						textboxNeededHeight,
						SWP_NOACTIVATE | SWP_NOZORDER);
	}

	if ( !resizeCommand.empty() )
		xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);
}

void Label::relayMouseMessageToBox(UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	pt.x = GET_X_LPARAM(lParam);
	pt.y = GET_Y_LPARAM(lParam);

	if (box)
	{
		MapWindowPoints(hWnd, box, &pt, 1);
		PostMessage(box, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
	else
	{
		MapWindowPoints(hWnd, GetLitestepDesktop(), &pt, 1);
		PostMessage(GetLitestepDesktop(), message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
}

void Label::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:

			POINT pt;
			GetCursorPos(&pt);
			
			RECT rc;
			GetWindowRect(hWnd, &rc);

			if(!PtInRect(&rc, pt))
			{
				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;
				if (!originalTextBackup.empty() && mouseInsideHREF)
				{
					current = currentBackup;
					setText(originalTextBackup);
				}
				mouseInsideHREF = false;
				onMouseLeave();	
			}

		break;

		case TIMER_UPDATE:

			if ( visible )
				update();

		break;
		
		case TIMER_SCROLL:

			int useSteps;
			if ( scrollStepsInitial > 0 )
				useSteps = scrollStepsInitial;
			else
				useSteps = scrollSteps;

			if ( visible && !frozen && (scroll == 2 || scroll == 4) && scrollPause > 0 && useSteps > 0 )
			{
				if ( scrollActualStep < useSteps )
				{
					scrollActualStep++;
				
					if (scroll == 2)
						scrollPosition -= scrollSpeed;
					else if (scroll == 4)
						scrollPosition += scrollSpeed;

					repaint();
				}
				else
				{
					scrollActualStep = 0;
					scrollStepsInitial = -1;
					frozen = true;
					SetTimer(hWnd, TIMER_SCROLLPAUSE, scrollPause, 0);
				}
			}
			//Normal Scroll (Horizontal or without Pause)
			else if ( visible && !frozen )
			{
				if (scroll == 1 || scroll == 2)
					scrollPosition -= scrollSpeed;
				else if (scroll == 3 || scroll == 4)
					scrollPosition += scrollSpeed;

				repaint();
			}

		break;

		case TIMER_AUTOHIDE:

			if ( visible )
				hide();
			else
				show();

		break;

		case TIMER_ALPHAFADEIN:

			if (UpdateAlpha <= alphaTransparency-alphaFadeStep)
			{
				setAlpha(UpdateAlpha, false);
				UpdateAlpha += alphaFadeStep;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				if (UpdateAlpha != alphaTransparency)
					setAlpha(alphaTransparency, false);
			}
		
		break;

		case TIMER_ALPHAFADEOUT:

			if (UpdateAlpha >= alphaFadeStep)
			{
				setAlpha(UpdateAlpha, false);
				UpdateAlpha -= alphaFadeStep;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				UpdateAlpha = 0;
				ShowWindow(hWnd, SW_HIDE);
			}

		break;

		case TIMER_FRAMEANIMATION:

			if ( visible )
			{
				if (hoverActive && useHover)
				{
					if (frameCounter == hoverimageFrameCount)
					{
						frameCounter = 0;
						loopCounter++;
					}
					if (loopCounter == hoverimageLoop)
					{
						frameCounter = -1;
						loopCounter = 0;
						KillTimer(hWnd, TIMER_FRAMEANIMATION);
					}
					
				}
				else
				{
					if (frameCounter == imageFrameCount)
					{
						frameCounter = 0;
						loopCounter++;
					}
					if (loopCounter == imageLoop)
					{
						frameCounter = -1;
						loopCounter = 0;
						KillTimer(hWnd, TIMER_FRAMEANIMATION);
					}
				}
				frameCounter++;

				repaint(true);
			}

		break;

		case TIMER_SCROLLPAUSE:

			if ( visible && frozen )
				frozen = false;

		break;
	}
}

bool Label::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	switch(message)
	{
		case LM_TEXTEDITBOX:
		{
			enterActionProcessing( (string) (const char *) lParam );
			return true;
		}

		case LM_SETLABELTEXT:
		{
			setText(string((const char *) lParam), true);
			return true;
		}

		case LM_SETCURSOR:
		{
			if (!cursor && !cursorPath.empty())
			{
				if (strstr(cursorPath.c_str(), ":\\") != NULL)
				{
					cursor = LoadCursorFromFile( cursorPath.c_str() );
				}
				else
				{	
					char imagePath[MAX_PATH_LENGTH] = {0};
					VarExpansion( imagePath, "$lsimagefolder$" );
					strcat(imagePath, cursorPath.c_str());

					cursor = LoadCursorFromFile( imagePath );
				}

				if (!cursor)
					cursor = LoadCursor( 0, IDC_ARROW );

				SetCursor( cursor );
			}
			else if (!cursor && cursorPath.empty())
			{
				cursor = LoadCursor( 0, IDC_ARROW );
				SetCursor( cursor );
			}
			else if (lParam != NULL)
			{
				SetCursor( LoadCursor(0, (const char *) lParam ) );
			}
			else
				SetCursor( cursor );

			return true;
		}

		case LM_EXTERNALDRAW:
		{
			if ( !parentIsAlphaMapped )
				parentIsAlphaMapped = true;

			if ( visible )
			{	
				HDC desthDC = (HDC) wParam;

				if ( alphaMap )
				{
					BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, alphaTransparency, AC_SRC_ALPHA };

					AlphaBlend(desthDC, actualx, actualy, actualwidth, actualheight, backgroundDC, 0, 0, actualwidth, actualheight, blendFunction);
				}
				else
				{
					TransparentBltLS(desthDC, actualx, actualy, actualwidth, actualheight, backupDC, 0, 0, RGB(255, 0, 255));
				}
			}
			return true;
		}

		case LM_EXTERNALDRAWNOTIFY:
		{
			if ( visible && (alphaMap || parentIsAlphaMapped) )
				repaint();

			return true;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DESTROY:
		{
			if(box)
			{
				if (TooltipHints)
				{
					removeHint(hWnd);
					DestroyWindow(TooltipHints);
					TooltipHints = 0;
				}

				if (hWndTextBox)
				{
					DestroyWindow(hWndTextBox);
					hWndTextBox = 0;
				}

				KillTimer(hWnd, TIMER_MOUSETRACK);
				KillTimer(hWnd, TIMER_UPDATE);
				KillTimer(hWnd, TIMER_SCROLL);
				KillTimer(hWnd, TIMER_AUTOHIDE);
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				KillTimer(hWnd, TIMER_FRAMEANIMATION);
				KillTimer(hWnd, TIMER_SCROLLPAUSE);

				box = NULL;
				hide();
			}
				
			hWnd = NULL;
			
			return false;
		}

		case WM_DROPFILES:
		{
			if (dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			int intFileAction = -1;

			char action[32] = {0};
			char szTarget[MAX_PATH_LENGTH] = {0};
			char *buffers[] = {action, szTarget};

			LCTokenize(dropCommand.c_str(), buffers, 2, NULL);

			if (stricmp(action, "move") == NULL)
				intFileAction = 1;
			else if (stricmp(action, "copy") == NULL)
				intFileAction = 2;
			else if (stricmp(action, "delete") == NULL)
				intFileAction = 0;
			else if (stricmp(action, "shortcut") == NULL)
				intFileAction = 3;
			else
				intFileAction = -1;

			if (intFileAction == -1)
			{
				int numDropped, i;
				char szFile[MAX_PATH_LENGTH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_PATH_LENGTH);
					if (szFile && szFile[0])
					{
						string fullCommand = (dropCommand + " \"" + szFile + "\""); // FIXME: ugly hack
						LSExecute(hWnd, fullCommand.c_str(), SW_SHOWNORMAL);
					}
				}

				DragFinish((HDROP) wParam);
				break;
			}

			string check = szTarget;
			if (check.substr(1, check.length()-1) == "\\")
				check.append( "*" );
			else
				check.append( "\\*" );
				
			//Test Path for Existence
			WIN32_FIND_DATA FindData;
			HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

			if (intFileAction >= 0 && intFileAction < 3)
			{
				char *szSourceFiles = NULL;
				char buffer[MAX_PATH_LENGTH];
				int numFiles;
				UINT len,scursor = 0;

				numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

				szSourceFiles = new char[MAX_PATH_LENGTH*numFiles+1];
				scursor = 0;

				for (int i=0;i<numFiles;i++)
				{
					len = DragQueryFile((HDROP) wParam, i, buffer, MAX_PATH_LENGTH);
					
					strcpy(szSourceFiles+scursor,buffer);
					scursor += len + 1;
				}
					
				DragFinish((HDROP) wParam);

				if( hFindHandle != INVALID_HANDLE_VALUE || intFileAction == 0 )
				{
					szSourceFiles[scursor] = 0;
					strcpy(buffer, szTarget);
					buffer[strlen(buffer)+1] = 0;
					
					SHFILEOPSTRUCT shfop;
					shfop.hwnd = NULL;
					if (intFileAction == 1)
						shfop.wFunc = FO_MOVE;
					else if (intFileAction == 2)
						shfop.wFunc = FO_COPY;
					else if (intFileAction == 0)
						shfop.wFunc = FO_DELETE;
					shfop.pFrom = szSourceFiles;
					shfop.pTo = buffer;
					shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
					SHFileOperation(&shfop);
				}
				
				delete[] szSourceFiles;
			}
			else if (hFindHandle != INVALID_HANDLE_VALUE && intFileAction == 3)
			{
				CString path = szTarget;
				if (path.Right(1) != "\\")
					path += "\\";
				int numDropped;
				char szFile[MAX_PATH_LENGTH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (int i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_PATH_LENGTH);

					HRESULT hres; 
					IShellLink* psl; 

					hres = CoCreateInstance(CLSID_ShellLink, NULL,
					CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID *) &psl); 
					if (SUCCEEDED(hres))
					{
						IPersistFile* ppf; 
						psl->SetPath(szFile); 
						psl->SetDescription("");
 
						hres = psl->QueryInterface(IID_IPersistFile, (LPVOID*)&ppf); 
						if (SUCCEEDED(hres))
						{
							CString work = szFile;
							work = work.Right(work.GetLength()-work.ReverseFind('\\')-1);
							work = work.Left(work.Find("."));
							work += ".lnk";

							WCHAR wsz[MAX_PATH]; 
							MultiByteToWideChar(CP_ACP, 0, (path + work), -1, wsz, MAX_PATH_LENGTH);  
							hres = ppf->Save(wsz, TRUE); 
							ppf->Release(); 
						}
						psl->Release(); 
					}
				}
				DragFinish((HDROP) wParam);
			}
			
			FindClose(hFindHandle);
			
			return false;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			if(labelInitialized && leftDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			if(labelInitialized && leftClickDownCommand.empty() && leftClickUpCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && labelHREFClickRegions.empty() && (!moveable || (moveable && moveButton != 1)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			if(labelInitialized && leftClickDownCommand.empty() && leftClickUpCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && labelHREFClickRegions.empty() && (!moveable || (moveable && moveButton != 1)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			if(labelInitialized && middleDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			if (focusOnEnter == 2)
				return true;

			if(labelInitialized && middleClickDownCommand.empty() && middleClickUpCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && (!moveable || (moveable && moveButton != 3)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			if (focusOnEnter == 2)
			{
				focusOnEnter = 1;
				return true;
			}

			if(labelInitialized && middleClickDownCommand.empty() && middleClickUpCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && (!moveable || (moveable && moveButton != 3)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));

			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			if(labelInitialized && rightDoubleClickCommand.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			if(labelInitialized && rightClickDownCommand.empty() && rightClickUpCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && (!moveable || (moveable && moveButton != 2)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			if(labelInitialized && rightClickDownCommand.empty() && rightClickUpCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && (!moveable || (moveable && moveButton != 2)) )
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MOUSEWHEEL:
		{
			POINT pt = {(int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam)};

			for(HWND hWndChild = ::GetTopWindow(hWnd); hWndChild != NULL; hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT))
			{
				ClientToScreen(hWnd, &pt);
				ScreenToClient(hWndChild, &pt);
				RECT childPosition;
				GetWindowRect(hWndChild, &childPosition);
				if ( PtInRect(&childPosition, pt) )
				{
					::SendMessage(hWndChild, WM_MOUSEWHEEL, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
					return true;
				}
			}

			if ((short)(HIWORD(wParam)) < 0)
			{
				onWheelDown(pt.x, pt.y);

				return true;
			}
			else
			{
				onWheelUp(pt.x, pt.y);

				return true;
			}
		}

		case WM_MOUSEMOVE:
		{
			if (moveable && moveKey != -1)
			{
				if (moveKey == 3)
					movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT));
				else if (moveKey == 2)
					movemodifierkeyPressed = (0x8000 & GetKeyState(VK_SHIFT));
				else if (moveKey == 1)
					movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
				else 
					movemodifierkeyPressed = (0x8000 & GetKeyState(VK_CONTROL));
			}
			else if (moveable && moveKey == -1)
				movemodifierkeyPressed = true;

			onMouseMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_ERASEBKGND:
		{
			return true;
		}

		case WM_CTLCOLOREDIT:
		{
			return true;
		}

		if ( (!alphaMap || !IsOS(OS_2KXP)) && !parentIsAlphaMapped )
		{
			case WM_PAINT:
			{	
				PAINTSTRUCT ps;
				HDC hDC;

				if(!wParam)
					hDC = BeginPaint(hWnd, &ps);
				else
					hDC = (HDC) wParam;

				onPaint(hDC);

				if(!wParam)
					EndPaint(hWnd, &ps);
		
				return true;	
			}
		}

		case WM_GETMINMAXINFO:
		{
			POINT maxsize = {autoMaxWidth, autoMaxHeight};
			POINT minsize = {autoMinWidth, autoMinHeight};

			((MINMAXINFO FAR*) lParam)->ptMaxTrackSize = maxsize;
			((MINMAXINFO FAR*) lParam)->ptMinTrackSize = minsize;

			return true;
		}

		case WM_WINDOWPOSCHANGING:
		{
			onWindowPosChanging((WINDOWPOS *) lParam);
			return true;
		}

		case WM_MOVE:
		{
			onMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_SIZE:
		{
			onSize((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		if ( useresizeBorder )
		{
			case WM_NCHITTEST:
			{
				POINT point;
				point.x = GET_X_LPARAM(lParam);
				point.y = GET_Y_LPARAM(lParam);
				ScreenToClient(hWnd, &point);

				lResult = NcHitTest(point.x, point.y);
			
				return lResult;
			}
		}
	}

	return false;
}

LRESULT Label::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Label *label = NULL;

	if(message == WM_NCCREATE)
	{
		label = (Label *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		label->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) label);
	}
	else
		label = (Label *) GetWindowLong(hWnd, 0);

	if(label)
	{
		LRESULT lResult = 0;

		if(label->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//------------------------------------------------------------
//Tooltips (Add, Update, Remove)
//------------------------------------------------------------

void Label::addHint(HWND hWnd, LPSTR caption)
{
	if (TooltipHints == 0)
		return;
	
	TOOLINFO ti;
	RECT clientRect;

	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;

	SendMessage(TooltipHints, TTM_ADDTOOL, 0, LPARAM(&ti));
}

void Label::updateHint(HWND hWnd, LPSTR caption, bool updatesize)
{
	if (TooltipHints == 0)
		return;	
	
	TOOLINFO ti;
	RECT clientRect;

	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;

	SendMessage(TooltipHints, TTM_UPDATETIPTEXT, 0, LPARAM(&ti));

	if (updatesize)
		SendMessage(TooltipHints, TTM_NEWTOOLRECT, 0, LPARAM(&ti));
}

void Label::removeHint(HWND hWnd)
{
	if (TooltipHints == 0)
		return;	
	
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;

	SendMessage(TooltipHints, TTM_DELTOOL, 0, LPARAM(&ti));
}

//------------------------------------------------------------
//Initialize and Update AlphaTransparency
//------------------------------------------------------------

void Label::UpdateLabelAlpha(int newAlphaTransparency)
{
	//Initialize LayeredWindow Function
	if (!alphaInitialized)
	{
		HMODULE hDLL = LoadLibrary ("user32");

		pUpdateLayeredWindow = 
					(PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

		pSetLayeredWindowAttributes = 
				(PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

		//Alpha Transparency Features Not Allowed
		if ( pUpdateLayeredWindow == NULL || pSetLayeredWindowAttributes == NULL )
		{
			alphaTransparency = 255;
			alphaMap = false;
			alphaFade = false;

			ModifyStyleEx(hWnd, WS_EX_LAYERED, 0 );

			layeredFlag = false;

			return;
		}

		alphaInitialized = true;
	}

	if (alphaMap)
	{
		//------------------------------------------------------------------
		RECT rectPosition;
		if ( box )
		{
			rectPosition.left = actualx;
			rectPosition.right = actualx+actualwidth;
			rectPosition.top = actualy;
			rectPosition.bottom = actualy+actualheight;
		}
		else
		{
			GetWindowRect(hWnd, &rectPosition);
		}
			
		int x = rectPosition.left;
		int y = rectPosition.top;
		int width = (rectPosition.right-rectPosition.left);
		int height = (rectPosition.bottom-rectPosition.top);

		if ( width <= 0 || height <= 0 )
			return;
			
		if(!backgroundBitmap)
		{
			if(!backgroundDC)
				backgroundDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=width;      // Set size you need
			alpha.bmiHeader.biHeight=height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			backgroundBitmap = CreateDIBSection( backgroundDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

			SelectObject(backgroundDC, backgroundBitmap);
		}

		//BG Painting
		if (background)
		{
			if ( (!normalAnim && !hoverAnim) || (usePressed && pressedActive) || frameCounter == 0)
			{
				if (hoverActive && useHover)
					background->apply(backgroundDC, 0, 0, width, height, 1, 1);
				else if (pressedActive && usePressed)
					background->apply(backgroundDC, 0, 0, width, height, 1, 2);
				else
					background->apply(backgroundDC, 0, 0, width, height, 1, 0);
			}
			else
			{
				if (hoverActive && useHover)
					background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
				else if (pressedActive && usePressed)
					background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
				else
					background->applyAnim(backgroundDC, 0, 0, width, height, 1, 0, frameCounter);
			}
		
			if ( !ghostlabelList.empty() )
			{
				for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
				{
					if ( (hoverActive && useHover && (*glit)->hover) || (pressedActive && usePressed && (*glit)->pressed) || (!hoverActive && !pressedActive && (*glit)->normal) )
					{
						if ( (*glit)->alphamap || (*glit)->alpha != 255 )
						{
							BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, (*glit)->alpha, AC_SRC_ALPHA };

							AlphaBlend(backgroundDC, (*glit)->x, (*glit)->y, ((*glit)->fullstretch ? width : (*glit)->width), ((*glit)->fullstretch ? height : (*glit)->height), (*glit)->memoryDC, 0, 0, (*glit)->width, (*glit)->height, blendFunction);
						}
						else
							TransparentBltLS(backgroundDC, (*glit)->x, (*glit)->y, ((*glit)->fullstretch ? width : (*glit)->width), ((*glit)->fullstretch ? height : (*glit)->height), (*glit)->memoryDC, 0, 0, RGB(255, 0, 255));
					}
				}
			}

			//Text Painting Function		
			if ( !text.empty() && !useTextFade )
				onTextPaint(backgroundDC, width, height);
			else if ( !text.empty() && useTextFade )
				TextFadeProcessing(backgroundDC, width, height);

			//Hooked Child Window AlphaMap Painting
			if ( hasHookedModules && (alphaMap || parentIsAlphaMapped) )
			{
				for(HWND hWndChild = ::GetTopWindow(hWnd); hWndChild != NULL; hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT))
					SendMessage(hWndChild, LM_EXTERNALDRAW, (long) backgroundDC, 0 );
			}

			//Prepare Combined Image
			if (labelInitialized)
			{
				BITMAP bm;
				GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

				BYTE *pPixel = (LPBYTE)bm.bmBits;

				for (int j = 0; j < bm.bmHeight * bm.bmWidth; ++j)
				{
					//Convert Text to Full Opacity
					if ( pPixel[3] == 0 )
						pPixel[3] = 255;

					//Magic Pink
					if ( hasHookedModules && pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255 )
					{
						pPixel[0] = 0;
						pPixel[1] = 0;
						pPixel[2] = 0;
						pPixel[3] = 0;
					}

					//Remove Prepared Alpha (1 back to 0)
					if (pPixel[3] == 1)
						pPixel[3] = 0;

					pPixel+= 4;
				}
			}	

			POINT ptWindowScreenPosition = {x, y};
			SIZE szWindow = {width, height};

			POINT ptSrc = {0,0};

			BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, newAlphaTransparency, AC_SRC_ALPHA };

			pUpdateLayeredWindow(hWnd, NULL, &ptWindowScreenPosition, &szWindow, backgroundDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );
		}
	//------------------------------------------------------------------
	}
	else
		pSetLayeredWindowAttributes(hWnd, 0, newAlphaTransparency, LWA_ALPHA);
}

void Label::TextFadeProcessing(HDC workingDC, int width, int height)
{
	if ( width <= 0 || height <= 0 )
		return;

	HDC	textFadeDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP textFadeBitmap = CreateDIBSection( textFadeDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	SelectObject(textFadeDC, textFadeBitmap);

	BitBlt(textFadeDC, 0, 0, width, height, workingDC, 0, 0, SRCCOPY);

	BITMAP bm;
	GetObject(textFadeBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	for (int j = 0; j < bm.bmHeight * bm.bmWidth; ++j)
	{
		if ( alphaMap )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 1;
		}
		else
		{
			pPixel[3] = 1;
		}

		pPixel+= 4;
	}

	onTextPaint(textFadeDC, width, height);

	GetObject(textFadeBitmap, sizeof(BITMAP), &bm);
	pPixel = (LPBYTE)bm.bmBits;

	int tmpx = 0;
	int tmpy = 0;
	int xval = 255;
	int yval = 255;
	for (j = 0; j < bm.bmHeight * bm.bmWidth; ++j)
	{
		//Convert Text to Full Opacity
		if (pPixel[3] == 0)
		{
			xval = 255;
			yval = 255;

			if ( useTextFade )
			{
				if (width != 0 )
				{
					tmpx = (j+1)%width;
					tmpy = (j+1)/width+1;

					if ( tmpx == 0 )
						tmpx = width-1;
				}
				else
				{
					tmpx = 1;
					tmpy = 1;
				}

				if ( tmpx < leftBorder+textFadeLeft )
				{
					if ( tmpx > leftBorder && textFadeLeft != 0 )
						xval = 255/textFadeLeft*(tmpx-leftBorder);
				}
				else if ( tmpx >= width-rightBorder-textFadeRight )
				{
					if ( tmpx <= width-rightBorder && textFadeRight != 0 )
						xval = 255/textFadeRight*(width-rightBorder-tmpx);
				}

				//Order is reverted Top<->Bottom
				if ( tmpy < bottomBorder+textFadeBottom )
				{
					if ( tmpy > bottomBorder && textFadeBottom != 0 )
						yval = 255/textFadeBottom*(tmpy-bottomBorder);
				}
				else if ( tmpy >= height-topBorder-textFadeTop )
				{
					if ( tmpy <= height-topBorder && textFadeTop != 0 )
						yval = 255/textFadeTop*(height-topBorder-tmpy);
				}
			}

			if ( xval == 255 )
				pPixel[3] = yval;
			else if ( yval == 255 )
				pPixel[3] = xval;
			else
				pPixel[3] = (xval+yval)/2;
				
			//PreMultiply
			pPixel[0] = pPixel[0] * pPixel[3] / 255;
			pPixel[1] = pPixel[1] * pPixel[3] / 255;
			pPixel[2] = pPixel[2] * pPixel[3] / 255;
		}
		else
		{
			pPixel[3] = 0; 
			if ( !alphaMap )
			{
				pPixel[0] = pPixel[0] * pPixel[3] / 255;
				pPixel[1] = pPixel[1] * pPixel[3] / 255;
				pPixel[2] = pPixel[2] * pPixel[3] / 255;
			}
		}
					
		pPixel+= 4;	
	}

	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, 255, AC_SRC_ALPHA };

	AlphaBlend(workingDC, 0, 0, width, height, textFadeDC, 0, 0, width, height, blendFunction);

	if(textFadeBitmap)
	{
		textFadeBitmap = (HBITMAP) SelectObject(textFadeDC, textFadeBitmap);
		DeleteObject(textFadeBitmap);
	}
	if (textFadeDC)
		DeleteDC(textFadeDC);
}

int Label::isCloseTo(POINT &pos, const POINT size, const int x, const int y, const int cx, const int cy, const int span)
{
	bool internalX = false;
	bool internalY = false;

	int tmpdockx = -1;
	int tmpdocky = -1;

	POINT tmppos = {pos.x, pos.y};

	if ( (tmppos.x > x-span) && (tmppos.x < x+span) )
	{
		if ( tmppos.y > y-span && tmppos.y < y+span )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
	}
	else if ( (tmppos.x > x+cx-span) && (tmppos.x < x+cx+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x+cx;
			internalX = true;
			tmpdockx = 2;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x+cx;
			internalX = true;
			tmpdockx = 2;
		}
	}
	else if ( (tmppos.x+size.x > x-span ) && (tmppos.x+size.x < x+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x-size.x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x-size.x;
			internalX = true;
			tmpdockx = 0;
		}
	}
	else if ( (tmppos.x+size.x > x+cx-span ) && (tmppos.x+size.x < x+cx+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x+cx-size.x;
			internalX = true;
			tmpdockx = 2;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x+cx-size.x;
			internalX = true;
			tmpdockx = 2;
		}
	}

	if ( (tmppos.y > y-span) && (tmppos.y < y+span) )
	{
		if ( tmppos.x > x-span && tmppos.x < x+span )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
	}
	else if ( tmppos.y > y+cy-span && tmppos.y < y+cy+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y+cy;
			internalY = true;
			tmpdocky = 3;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx  
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y+cy;
			internalY = true;
			tmpdocky = 3;
		}
	}
	else if ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y-size.y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y-size.y;
			internalY = true;
			tmpdocky = 1;
		}
	}
	else if ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y+cy-size.y;
			internalY = true;
			tmpdocky = 3;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y+cy-size.y;
			internalY = true;
			tmpdocky = 3;
		}
	}

	if (!internalX && !internalY)
		return 0;
	else if (internalX && !internalY)
	{
		dockedtoSide = tmpdockx;
		return 1;
	}
	else if (!internalX && internalY)
	{
		dockedtoSide = tmpdocky;
		return 2;
	}
	else
	{
		if ( tmpdockx == 0 && tmpdocky == 1 )
			dockedtoSide = 4;
		else if ( tmpdockx == 2 && tmpdocky == 1 )
			dockedtoSide = 5;
		else if ( tmpdockx == 0 && tmpdocky == 3 )
			dockedtoSide = 6;
		else if ( tmpdockx == 2 && tmpdocky == 3 )
			dockedtoSide = 7;
		return 3;
	}
}

void Label::ontemplateButtonDown(int actualMouseClick, const string &clickDownCommand)
{
	//Stop Animation Timer (Normal & Hover)
	KillTimer(hWnd, TIMER_FRAMEANIMATION);

	if(moveable && moveButton == actualMouseClick)
	{
		if(!mousePressed && !movemodifierkeyPressed)
			mousePressed = true;
		else if (!mousePressed && movemodifierkeyPressed)
		{
			mousePressed = true;
			moveButtonPressed = actualMouseClick;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);	
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	if (usePressed)
	{
		hoverActive = false;
		pressedActive = true;
		repaint(true);
	}
	
	SetCapture(hWnd);
	
	if( !clickDownCommand.empty() )
		xLSExecute(hWnd, clickDownCommand.c_str(), SW_SHOWNORMAL);
}

void Label::ontemplateButtonUp(const int &x, const int &y, int actualMouseClick, const StringList &labelClickRegions, const string &clickCommand, const string &clickUpCommand)
{
	if ( moveable && moveButton == actualMouseClick )
	{	
		//To be on the save side
		ReleaseCapture();

		if ( mousePressed )
		{
			SendMessage(hWnd, LM_SETCURSOR, 0, 0);
		
			if ( !checkRegion(x, y, labelClickRegions) ) 
			{
				if (actualMouseClick == 1)
					checkRegion(x, y, labelHREFClickRegions, true);
			}
			
			mousePressed = false;
			moveButtonPressed = 0;

			//Docking
			if ( hasMoved )
			{
				for(StringListIterator sit = snapToList.begin(); sit != snapToList.end(); sit++)
				{
					ParseBangCommand(hWnd, "!labeldock", (*sit).c_str());
				}
				snapToList.clear();
			}

			if(hasMoved && labelInitialized && !moveCommand.empty())
				xLSExecute(hWnd, moveCommand.c_str(), SW_SHOWNORMAL);
			
			hasMoved = false;
		}
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();

		if( mousePressed )
		{
			if ( !checkRegion(x, y, labelClickRegions) )
			{
				if ( actualMouseClick == 1)
					checkRegion(x, y, labelHREFClickRegions, true);
			}

			mousePressed = false;
			moveButtonPressed = 0;
		}
	}

	if (usePressed)
	{
		hoverActive = true;
		pressedActive = false;
		repaint(true);
	}

	//Animation RESTART
	//Restart Hover Anim
	if (useHover && hoverAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
	}
	//Restart Normal Anim
	else if (normalAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}

	if( !clickCommand.empty() )
	{
		if ( hWndTextBox && (stricmp(clickCommand.c_str(), ".editbox") == NULL || stricmp(clickCommand.c_str(), "\".editbox\"") == NULL) )
		{
			ShowWindow(hWndTextBox, SW_SHOW);
			textboxVisible = true;
			SetFocus(hWndTextBox);
		}
		else
		{
			xLSExecute(hWnd, clickCommand.c_str(), SW_SHOWNORMAL);
		}
	}

	if( !clickUpCommand.empty() )
		xLSExecute(hWnd, clickUpCommand.c_str(), SW_SHOWNORMAL);
}

void Label::enterActionProcessing(string input)
{
	if ( !input.empty() )
	{
		if ( stricmp(enterAction.c_str(), ".execute") == NULL || stricmp(enterAction.c_str(), "\".execute\"") == NULL )
		{
			xLSExecute(hWnd, input.c_str(), SW_SHOWNORMAL);
		}
		else if  ( stricmp(enterAction.c_str(), ".settext") == NULL || stricmp(enterAction.c_str(), "\".settext\"") == NULL )
		{
			SendMessage(hWnd, LM_SETLABELTEXT, 0, (long) input.c_str());
		}
		else
		{
			CString worker = enterAction.c_str();
			worker.Replace("%[textinput]%", input.c_str());
			xLSExecute(hWnd, worker, SW_SHOWNORMAL);
		}

		SetWindowText(hWndTextBox,"");
	}
			
	if ( textboxAutoHide )
	{
		ShowWindow(hWndTextBox, SW_HIDE);
		textboxVisible = false;
	}
}
